#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import threading
import time

import jwt
from bravado.exception import HTTPUnauthorized
from bravado.requests_client import Authenticator
from oauthlib.oauth2 import (
    OAuth2Error,
    TokenExpiredError,
)
from requests.auth import AuthBase
from requests_oauthlib import OAuth2Session

from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.common.exceptions import NeptuneInvalidApiTokenException
from neptune.common.utils import update_session_proxies

_decoding_options = {
    "verify_signature": False,
    "verify_exp": False,
    "verify_nbf": False,
    "verify_iat": False,
    "verify_aud": False,
    "verify_iss": False,
}


class NeptuneAuth(AuthBase):
    __LOCK = threading.RLock()

    def __init__(self, session_factory):
        self.session_factory = session_factory
        self.session = session_factory()
        self.token_expires_at = 0

    def __call__(self, r):
        try:
            return self._add_token(r)
        except TokenExpiredError:
            self._refresh_token()
            return self._add_token(r)

    def _add_token(self, r):
        r.url, r.headers, r.body = self.session._client.add_token(
            r.url, http_method=r.method, body=r.body, headers=r.headers
        )
        return r

    @with_api_exceptions_handler
    def refresh_token_if_needed(self):
        if self.token_expires_at - time.time() < 30:
            self._refresh_token()

    def _refresh_token(self):
        with self.__LOCK:
            try:
                self._refresh_session_token()
            except OAuth2Error:
                # for some reason oauth session is no longer valid. Retry by creating new fresh session
                # we can safely ignore this error, as it will be thrown again if it's persistent
                try:
                    self.session.close()
                except Exception:
                    pass
                self.session = self.session_factory()
                self._refresh_session_token()

    def _refresh_session_token(self):
        self.session.refresh_token(self.session.auto_refresh_url, verify=self.session.verify)
        if self.session.token is not None and self.session.token.get("access_token") is not None:
            decoded_json_token = jwt.decode(self.session.token.get("access_token"), options=_decoding_options)
            self.token_expires_at = decoded_json_token.get("exp")


class NeptuneAuthenticator(Authenticator):
    def __init__(self, api_token, backend_client, ssl_verify, proxies):
        super(NeptuneAuthenticator, self).__init__(host="")

        # We need to pass a lambda to be able to re-create fresh session at any time when needed
        def session_factory():
            try:
                auth_tokens = backend_client.api.exchangeApiToken(X_Neptune_Api_Token=api_token).response().result
            except HTTPUnauthorized:
                raise NeptuneInvalidApiTokenException()

            decoded_json_token = jwt.decode(auth_tokens.accessToken, options=_decoding_options)
            expires_at = decoded_json_token.get("exp")
            client_name = decoded_json_token.get("azp")
            refresh_url = "{realm_url}/protocol/openid-connect/token".format(realm_url=decoded_json_token.get("iss"))
            token = {
                "access_token": auth_tokens.accessToken,
                "refresh_token": auth_tokens.refreshToken,
                "expires_in": expires_at - time.time(),
            }

            session = OAuth2Session(
                client_id=client_name,
                token=token,
                auto_refresh_url=refresh_url,
                auto_refresh_kwargs={"client_id": client_name},
                token_updater=_no_token_updater,
            )
            session.verify = ssl_verify

            update_session_proxies(session, proxies)
            return session

        self.auth = NeptuneAuth(session_factory)

    def matches(self, url):
        return True

    def apply(self, request):
        self.auth.refresh_token_if_needed()
        request.auth = self.auth
        return request


def _no_token_updater():
    # For unit tests.
    return None



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StringSeries"]

import time
from itertools import cycle
from typing import (
    TYPE_CHECKING,
    Iterable,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from neptune.internal.types.stringify_value import StringifyValue
from neptune.internal.utils import (
    is_collection,
    is_stringify_value,
)
from neptune.types.series.series import Series

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")

MAX_STRING_SERIES_VALUE_LENGTH = 1000


class StringSeries(Series):
    def __init__(
        self,
        values: Union[Iterable[str], StringifyValue],
        timestamps: Optional[Sequence[float]] = None,
        steps: Optional[Sequence[float]] = None,
    ):
        if is_stringify_value(values):
            values = list(map(str, values.value))

        if not is_collection(values):
            raise TypeError("`values` is not a collection")

        self._truncated = any([len(value) > MAX_STRING_SERIES_VALUE_LENGTH for value in values])
        self._values = [value[:MAX_STRING_SERIES_VALUE_LENGTH] for value in values]

        if steps is None:
            self._steps = cycle([None])
        else:
            assert len(values) == len(steps)
            self._steps = steps

        if timestamps is None:
            self._timestamps = cycle([time.time()])
        else:
            assert len(values) == len(timestamps)
            self._timestamps = timestamps

    @property
    def steps(self):
        return self._steps

    @property
    def timestamps(self):
        return self._timestamps

    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        return visitor.visit_string_series(self)

    @property
    def values(self):
        return self._values

    @property
    def truncated(self):
        """True if any value had to be truncated to `MAX_STRING_SERIES_VALUE_LENGTH`"""
        return self._truncated

    def __str__(self):
        return "StringSeries({})".format(str(self.values))



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_backend"]

from typing import Optional

from neptune.internal.credentials import Credentials
from neptune.types.mode import Mode

from .hosted_neptune_backend import HostedNeptuneBackend
from .neptune_backend import NeptuneBackend
from .neptune_backend_mock import NeptuneBackendMock
from .offline_neptune_backend import OfflineNeptuneBackend


def get_backend(mode: Mode, api_token: Optional[str] = None, proxies: Optional[dict] = None) -> NeptuneBackend:
    if mode == Mode.ASYNC:
        return HostedNeptuneBackend(credentials=Credentials.from_token(api_token=api_token), proxies=proxies)
    elif mode == Mode.SYNC:
        return HostedNeptuneBackend(credentials=Credentials.from_token(api_token=api_token), proxies=proxies)
    elif mode == Mode.DEBUG:
        return NeptuneBackendMock()
    elif mode == Mode.OFFLINE:
        return OfflineNeptuneBackend()
    elif mode == Mode.READ_ONLY:
        return HostedNeptuneBackend(credentials=Credentials.from_token(api_token=api_token), proxies=proxies)
    else:
        raise ValueError(f"mode should be one of {[m for m in Mode]}")



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class GitInfo(object):
    """Class that keeps information about a git repository in experiment.

    When :meth:`~neptune.projects.Project.create_experiment` is invoked, instance of this class is created to
    store information about git repository.
    This information is later presented in the experiment details tab in the Neptune web application.

    Args:
        commit_id (:obj:`str`): commit id sha.
        message (:obj:`str`, optional, default is ``""``): commit message.
        author_name (:obj:`str`, optional, default is ``""``): commit author username.
        author_email (:obj:`str`, optional, default is ``""``): commit author email.
        commit_date (:obj:`datetime.datetime`, optional, default is ``""``): commit datetime.
        repository_dirty (:obj:`bool`, optional, default is ``True``):
            ``True``, if the repository has uncommitted changes, ``False`` otherwise.
    """

    def __init__(
        self,
        commit_id,
        message="",
        author_name="",
        author_email="",
        commit_date="",
        repository_dirty=True,
        active_branch="",
        remote_urls=None,
    ):
        if remote_urls is None:
            remote_urls = []
        if commit_id is None:
            raise TypeError("commit_id must not be None")

        self.commit_id = commit_id
        self.message = message
        self.author_name = author_name
        self.author_email = author_email
        self.commit_date = commit_date
        self.repository_dirty = repository_dirty
        self.active_branch = active_branch
        self.remote_urls = remote_urls

    def __eq__(self, o):
        return o is not None and self.__dict__ == o.__dict__

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return "GitInfo({})".format(self.commit_id)

    def __repr__(self):
        return str(self)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# flake8: noqa
from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.common.utils import (
    IS_MACOS,
    IS_WINDOWS,
    NoopObject,
    align_channels_on_x,
    as_list,
    assure_directory_exists,
    assure_project_qualified_name,
    discover_git_repo_location,
    file_contains,
    get_channel_name_stems,
    get_git_info,
    glob,
    in_docker,
    is_float,
    is_ipython,
    is_nan_or_inf,
    is_notebook,
    map_keys,
    map_values,
    merge_dataframes,
    update_session_proxies,
    validate_notebook_path,
)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import platform

from neptune.common.envs import (
    API_TOKEN_ENV_NAME,
    PROJECT_ENV_NAME,
)

UNIX_STYLES = {
    "h1": "\033[95m",
    "h2": "\033[94m",
    "blue": "\033[94m",
    "python": "\033[96m",
    "bash": "\033[95m",
    "warning": "\033[93m",
    "correct": "\033[92m",
    "fail": "\033[91m",
    "bold": "\033[1m",
    "underline": "\033[4m",
    "end": "\033[0m",
}

WINDOWS_STYLES = {
    "h1": "",
    "h2": "",
    "python": "",
    "bash": "",
    "warning": "",
    "correct": "",
    "fail": "",
    "bold": "",
    "underline": "",
    "end": "",
}

EMPTY_STYLES = {
    "h1": "",
    "h2": "",
    "python": "",
    "bash": "",
    "warning": "",
    "correct": "",
    "fail": "",
    "bold": "",
    "underline": "",
    "end": "",
}

if platform.system() in ["Linux", "Darwin"]:
    STYLES = UNIX_STYLES
elif platform.system() == "Windows":
    STYLES = WINDOWS_STYLES
else:
    STYLES = EMPTY_STYLES


class NeptuneException(Exception):
    def __eq__(self, other):
        if type(other) is type(self):
            return super().__eq__(other) and str(self).__eq__(str(other))
        else:
            return False

    def __hash__(self):
        return hash((super().__hash__(), str(self)))


class NeptuneInvalidApiTokenException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneInvalidApiTokenException------------------------------------------------
{end}
The provided API token is invalid.
Make sure you copied and provided your API token correctly.

You can get it or check if it is correct here:
    - https://app.neptune.ai/get_my_api_token

There are two options to add it:
    - specify it in your code
    - set as an environment variable in your operating system.

{h2}CODE{end}
Pass the token to the {bold}init_run(){end} function via the {bold}api_token{end} argument:
    {python}neptune.init_run(project='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}="YOUR_API_TOKEN"{end}

    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}="YOUR_API_TOKEN"{end}

and skip the {bold}api_token{end} argument of the {bold}init_run(){end} function:
    {python}neptune.init_run(project='WORKSPACE_NAME/PROJECT_NAME'){end}

You may also want to check the following docs page:
    - https://docs.neptune.ai/setup/setting_api_token/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(env_api_token=API_TOKEN_ENV_NAME, **STYLES))


class UploadedFileChanged(NeptuneException):
    def __init__(self, filename: str):
        super().__init__("File {} changed during upload, restarting upload.".format(filename))


class InternalClientError(NeptuneException):
    def __init__(self, msg: str):
        message = """
{h1}
----InternalClientError-----------------------------------------------------------------------
{end}
The Neptune client library encountered an unexpected internal error:
{msg}

Please contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(msg=msg, **STYLES))


class ClientHttpError(NeptuneException):
    def __init__(self, status, response):
        self.status = status
        self.response = response
        message = """
{h1}
----ClientHttpError-----------------------------------------------------------------------
{end}
The Neptune server returned the status {fail}{status}{end}.

The server response was:
{fail}{response}{end}

Verify the correctness of your call or contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(status=status, response=response, **STYLES))


class NeptuneApiException(NeptuneException):
    pass


class Forbidden(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Forbidden-----------------------------------------------------------------------
{end}
You don't have permission to access the given resource.

    - Verify that your API token is correct.
      See: https://app.neptune.ai/get_my_api_token

    - Verify that the provided project name is correct.
      The correct project name should look like this: {correct}WORKSPACE_NAME/PROJECT_NAME{end}
      It has two parts:
          - {correct}WORKSPACE_NAME{end}: can be your username or your organization name
          - {correct}PROJECT_NAME{end}: the name specified for the project

   - Ask your organization administrator to grant you necessary privileges to the project.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class Unauthorized(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Unauthorized-----------------------------------------------------------------------
{end}
You don't have permission to access the given resource.

    - Verify that your API token is correct.
      See: https://app.neptune.ai/get_my_api_token

    - Verify that the provided project name is correct.
      The correct project name should look like this: {correct}WORKSPACE_NAME/PROJECT_NAME{end}
      It has two parts:
          - {correct}WORKSPACE_NAME{end}: can be your username or your organization name
          - {correct}PROJECT_NAME{end}: the name specified for the project

   - Ask your organization administrator to grant you necessary privileges to the project.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class NeptuneAuthTokenExpired(Unauthorized):
    def __init__(self):
        super().__init__("Authorization token expired")


class InternalServerError(NeptuneApiException):
    def __init__(self, response):
        message = """
{h1}
----InternalServerError-----------------------------------------------------------------------
{end}
The Neptune client library encountered an unexpected internal server error.

The server response was:
{fail}{response}{end}

Please try again later or contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(response=response, **STYLES))


class NeptuneConnectionLostException(NeptuneException):
    def __init__(self, cause: Exception):
        self.cause = cause
        message = """
{h1}
----NeptuneConnectionLostException---------------------------------------------------------
{end}
The connection to the Neptune server was lost.
If you are using the asynchronous (default) connection mode, Neptune continues to locally track your metadata and continuously tries to re-establish a connection to the Neptune servers.
If the connection is not re-established, you can upload your data later with the Neptune Command Line Interface tool:
    {bash}neptune sync -p workspace_name/project_name{end}

What should I do?
    - Check if your computer is connected to the internet.
    - If your connection is unstable, consider working in offline mode:
        {python}run = neptune.init_run(mode="offline"){end}

You can find detailed instructions on the following doc pages:
    - https://docs.neptune.ai/api/connection_modes/#offline-mode
    - https://docs.neptune.ai/api/neptune_sync/

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/connection_modes/#connectivity-issues

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(message.format(**STYLES))


class NeptuneSSLVerificationError(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneSSLVerificationError-----------------------------------------------------------------------
{end}

The Neptune client was unable to verify your SSL Certificate.

{bold}What could have gone wrong?{end}
    - You are behind a proxy that inspects traffic to Neptune servers.
        - Contact your network administrator
    - The SSL/TLS certificate of your on-premises installation is not recognized due to a custom Certificate Authority (CA).
        - To check, run the following command in your terminal:
            {bash}curl https://<your_domain>/api/backend/echo {end}
        - Where <your_domain> is the address that you use to access Neptune app, such as abc.com
        - Contact your network administrator if you get the following output:
            {fail}"curl: (60) server certificate verification failed..."{end}
    - Your machine software is outdated.
        - Minimal OS requirements:
            - Windows >= XP SP3
            - macOS >= 10.12.1
            - Ubuntu >= 12.04
            - Debian >= 8

{bold}What can I do?{end}
You can manually configure Neptune to skip all SSL checks. To do that,
set the NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE environment variable to 'TRUE'.
{bold}Note: This might mean that your connection is less secure{end}.

Linux/Unix
In your terminal run:
    {bash}export NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

Windows
In your terminal run:
    {bash}set NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

Jupyter notebook
In your code cell:
    {bash}%env NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/environment_variables/#neptune_allow_self_signed_certificate


{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(message.format(**STYLES))


class FileNotFound(NeptuneException):
    def __init__(self, path):
        super(FileNotFound, self).__init__("File {} doesn't exist.".format(path))


class InvalidNotebookPath(NeptuneException):
    def __init__(self, path):
        super(InvalidNotebookPath, self).__init__(
            "File {} is not a valid notebook. Should end with .ipynb.".format(path)
        )


class NeptuneIncorrectProjectQualifiedNameException(NeptuneException):
    def __init__(self, project_qualified_name):
        message = """
{h1}
----NeptuneIncorrectProjectQualifiedNameException-----------------------------------------------------------------------
{end}
Project qualified name {fail}"{project_qualified_name}"{end} you specified was incorrect.

The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
It has two parts:
    - {correct}WORKSPACE{end}: which can be your username or your organization name
    - {correct}PROJECT_NAME{end}: which is the actual project name you chose

For example, a project {correct}neptune-ai/credit-default-prediction{end} parts are:
    - {correct}neptune-ai{end}: {underline}WORKSPACE{end} our company organization name
    - {correct}credit-default-prediction{end}: {underline}PROJECT_NAME{end} a project name

The URL to this project looks like this: https://app.neptune.ai/neptune-ai/credit-default-prediction

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/workspace-project-and-user-management/index.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneIncorrectProjectQualifiedNameException, self).__init__(
            message.format(project_qualified_name=project_qualified_name, **STYLES)
        )


class NeptuneMissingProjectQualifiedNameException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneMissingProjectQualifiedNameException-------------------------------------------------------------------------
{end}
Neptune client couldn't find your project name.

There are two options two add it:
    - specify it in your code
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass it to {bold}neptune.init(){end} via {bold}project_qualified_name{end} argument:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end}
or export or set an environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In your terminal run:
       {bash}export {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

    {correct}Windows{end}
    In your CMD run:
       {bash}set {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

and skip the {bold}project_qualified_name{end} argument of {bold}neptune.init(){end}:
    {python}neptune.init(api_token='YOUR_API_TOKEN'){end}

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/workspace-project-and-user-management/index.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(NeptuneMissingProjectQualifiedNameException, self).__init__(
            message.format(env_project=PROJECT_ENV_NAME, **STYLES)
        )


class NotAFile(NeptuneException):
    def __init__(self, path):
        super(NotAFile, self).__init__("Path {} is not a file.".format(path))


class NotADirectory(NeptuneException):
    def __init__(self, path):
        super(NotADirectory, self).__init__("Path {} is not a directory.".format(path))


class WritingToArchivedProjectException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----WritingToArchivedProjectException-----------------------------------------------------------------------
{end}
You're trying to write to a project that was archived.

Set the project as active again or use mode="read-only" at initialization to fetch metadata from it.

{correct}Need help?{end}-> https://docs.neptune.ai/help/error_writing_to_archived_project/
"""
        super(WritingToArchivedProjectException, self).__init__(message.format(**STYLES))



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The management API lets you perform various neptune.ai administration actions.

- Create and delete projects
- List projects you can access
- Add and remove project members
- List members of projects and workspaces

Most actions can also be performed by service accounts.

Functions:
    get_project_list()
    create_project()
    delete_project()
    get_project_member_list()
    add_project_member()
    remove_project_member()
    get_workspace_member_list()
    add_project_service_account()
    remove_project_service_account()
    get_project_service_account_list()
    get_workspace_service_account_list()
    trash_objects()

Usage examples
--------------

Getting projects in a workspace as a list:
>>> projects = management.get_project_list()

Creating a new project:
>>> management.create_project(
...     name="ml-team/classification",
...     key="CLS",
...     visibility="workspace",
... )

Deleting a project:
>>> management.delete_project(project="ml-team/classification")

Get project members list as dictionary, with usernames as keys and roles as values:
>>> members = management.get_project_member_list(project="ml-team/classification")

Assign a user to a project and specify a role:
>>> management.add_project_member(
...    project="ml-team/classification", username="jackie", role="contributor"
... )

Remove a user from a project:
>>> management.remove_project_member(project="ml-team/classification", username="janus")

Get workspace members list as dictionary, with usernames as keys and roles as values:
>>> management.get_workspace_member_list(workspace="ml-team")

Assign service account to project:
>>> management.add_project_service_account(
...     project="ml-team/classification",
...     service_account_name="cicd@ml-team",
...     role="contributor",
... )

Remove service account from project:
>>> management.remove_project_service_account(
...     project="ml-team/classification", service_account_name="cicd@ml-team"
... )

Get list of project service accounts as dictionary, with usernames as keys and roles as values:
>>> management.get_project_service_account_list(project="ml-team/classification")

Get list of workspace service accounts as dictionary, with usernames as keys and roles as values:
>>> management.get_workspace_service_account_list(workspace="ml-team")

Move one or more Neptune objects to the project trash:
>>> project_name = "ml-team/classification"
>>> # Connect to your project:
... project = neptune.init_project(project=project_name)
>>> # Fetch runs tagged as "trash":
... runs_to_trash_df = project.fetch_runs_table(tag="trash").to_pandas()
>>> # Turn run IDs into a list:
... runs_to_trash = runs_to_trash_df["sys/id"].tolist()
>>> # Move the runs to trash:
... management.trash_objects(project=project_name, ids=runs_to_trash)

---

See also the API reference in the docs: https://docs.neptune.ai/api/management
"""
from .internal.api import (
    WorkspaceMemberRole,
    add_project_member,
    add_project_service_account,
    create_project,
    delete_project,
    get_project_list,
    get_project_member_list,
    get_project_service_account_list,
    get_workspace_member_list,
    get_workspace_service_account_list,
    invite_to_workspace,
    remove_project_member,
    remove_project_service_account,
    trash_objects,
)
from .internal.types import (
    MemberRole,
    ProjectVisibility,
)

__all__ = [
    "get_project_list",
    "create_project",
    "delete_project",
    "get_project_member_list",
    "add_project_member",
    "remove_project_member",
    "get_workspace_member_list",
    "invite_to_workspace",
    "WorkspaceMemberRole",
    "add_project_service_account",
    "remove_project_service_account",
    "get_project_service_account_list",
    "get_workspace_service_account_list",
    "trash_objects",
    "MemberRole",
    "ProjectVisibility",
]



# MIT License
#
# Copyright (c) 1990-2022 Robert Nowotny
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# flake8: noqa

import pathlib
import sys

import __main__  # noqa

__all__ = ["empty_path", "get_path_executed_script"]

empty_path = pathlib.Path()


def get_path_executed_script() -> pathlib.Path:
    """
    getting the full path of the program from which a Python module is running

    >>> ### TEST get it via __main__.__file__
    >>> # Setup
    >>> # force __main__.__file__ valid
    >>> save_main_file = str(__main__.__file__)
    >>> __main__.__file__ = __file__

    >>> # Test via __main__.__file__
    >>> assert get_path_executed_script() == pathlib.Path(__file__).resolve()


    >>> ### TEST get it via sys.argv
    >>> # Setup
    >>> # force __main__.__file__ invalid
    >>> __main__.__file__ = str((pathlib.Path(__file__).parent / 'invalid_file.py'))  # .resolve() seems not to work on a non existing file in python 3.5

    >>> # force sys.argv valid
    >>> save_sys_argv = list(sys.argv)
    >>> valid_path = str((pathlib.Path(__file__).resolve()))
    >>> sys.argv = [valid_path]

    >>> # Test via sys.argv
    >>> assert get_path_executed_script() == pathlib.Path(__file__).resolve()


    >>> ### TEST get it via stack
    >>> # Setup
    >>> # force sys.argv invalid
    >>> invalid_path = str((pathlib.Path(__file__).parent / 'invalid_file.py'))  # .resolve() seems not to work on a non existing file in python 3.5
    >>> sys.argv = [invalid_path]


    >>> assert get_path_executed_script()

    >>> # teardown
    >>> __main__.__file__ = save_main_file
    >>> sys.argv = list(save_sys_argv)

    """

    # try to get it from __main__.__file__ - does not work under pytest, doctest
    path_candidate = get_fullpath_from_main_file()
    if path_candidate != empty_path:
        return path_candidate

    # try to get it from sys_argv - does not work when loaded from uwsgi, works in eclipse and pydev
    path_candidate = get_fullpath_from_sys_argv()
    if path_candidate != empty_path:
        return path_candidate

    return empty_path


def get_fullpath_from_main_file() -> pathlib.Path:
    """try to get it from __main__.__file__ - does not work under pytest, doctest

    >>> # test no attrib __main__.__file__
    >>> save_main_file = str(__main__.__file__)
    >>> delattr(__main__, '__file__')
    >>> assert get_fullpath_from_main_file() == empty_path
    >>> setattr(__main__, '__file__', save_main_file)

    """
    if not hasattr(sys.modules["__main__"], "__file__"):
        return empty_path

    arg_string = str(sys.modules["__main__"].__file__)
    valid_executable_path = get_valid_executable_path_or_empty_path(arg_string)
    return valid_executable_path


def get_fullpath_from_sys_argv() -> pathlib.Path:
    """try to get it from sys_argv - does not work when loaded from uwsgi, works in eclipse and pydev

    >>> # force test invalid sys.path
    >>> save_sys_argv = list(sys.argv)
    >>> invalid_path = str((pathlib.Path(__file__).parent / 'invalid_file.py'))  # .resolve() seems not to work on a non existing file in python 3.5
    >>> sys.argv = [invalid_path]
    >>> assert get_fullpath_from_sys_argv() == pathlib.Path()
    >>> sys.argv = list(save_sys_argv)

    >>> # force test valid sys.path
    >>> save_sys_path = list(sys.argv)
    >>> valid_path = str((pathlib.Path(__file__).resolve()))
    >>> sys.argv = [valid_path]
    >>> assert get_fullpath_from_sys_argv() == pathlib.Path(valid_path)
    >>> sys.argv = list(save_sys_argv)


    """

    for arg_string in sys.argv:
        valid_executable_path = get_valid_executable_path_or_empty_path(arg_string)
        if valid_executable_path != empty_path:
            return valid_executable_path
    return empty_path


def get_valid_executable_path_or_empty_path(arg_string: str) -> pathlib.Path:
    arg_string = remove_doctest_and_docrunner_parameters(arg_string)
    arg_string = add_python_extension_if_not_there(arg_string)
    path = pathlib.Path(arg_string)
    try:
        if path.is_file():
            path = path.resolve()  # .resolve does not work on a non existing file in python 3.5
            return path
        else:
            return empty_path
    except Exception as e:
        return empty_path


def remove_doctest_and_docrunner_parameters(arg_string: str) -> str:
    """
    >>> # Setup
    >>> arg_string_with_parameter = __file__ + '::::::some docrunner parameter'
    >>> arg_string_without_parameter = __file__

    >>> # Test with and without docrunner parameters
    >>> assert remove_doctest_and_docrunner_parameters(arg_string_with_parameter) == __file__
    >>> assert remove_doctest_and_docrunner_parameters(arg_string_without_parameter) == __file__
    """
    path = arg_string.split("::", 1)[0]
    return path


def add_python_extension_if_not_there(arg_string: str) -> str:
    """
    >>> # Setup
    >>> arg_string_with_py = __file__
    >>> arg_string_without_py = __file__.rsplit('.py',1)[0]

    >>> # Test with and without .py suffix
    >>> assert add_python_extension_if_not_there(arg_string_with_py) == __file__
    >>> assert add_python_extension_if_not_there(arg_string_without_py) == __file__

    """

    if not arg_string.endswith(".py"):
        arg_string = arg_string + ".py"
    return arg_string



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["handle_json_errors"]

from typing import (
    Any,
    Callable,
    Dict,
    Optional,
)


def handle_json_errors(
    content: Dict[str, Any],
    source_exception: Exception,
    error_processors: Dict[str, Callable[[Dict[str, Any]], Exception]],
    default_exception: Optional[Exception] = None,
) -> None:
    error_type: Optional[str] = content.get("errorType")
    error_processor = error_processors.get(error_type)

    if error_processor:
        raise error_processor(content) from source_exception
    elif default_exception:
        raise default_exception from source_exception
    raise source_exception



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["RunState"]

import enum

from neptune.common.exceptions import NeptuneException


class RunState(enum.Enum):
    active = "Active"
    inactive = "Inactive"

    _api_active = "running"
    _api_inactive = "idle"

    @classmethod
    def from_string(cls, value: str) -> "RunState":
        try:
            return cls(value.capitalize())
        except ValueError as e:
            raise NeptuneException(f"Can't map RunState to API: {value}") from e

    @staticmethod
    def from_api(value: str) -> "RunState":
        if value == RunState._api_active.value:
            return RunState.active
        elif value == RunState._api_inactive.value:
            return RunState.inactive
        else:
            raise NeptuneException(f"Unknown RunState: {value}")

    def to_api(self) -> str:
        if self is RunState.active:
            return self._api_active.value
        if self is RunState.inactive:
            return self._api_inactive.value



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Datetime"]

import typing
from datetime import datetime

from neptune.attributes.atoms.copiable_atom import CopiableAtom
from neptune.internal.container_type import ContainerType
from neptune.internal.operation import AssignDatetime
from neptune.internal.types.stringify_value import StringifyValue
from neptune.internal.utils import verify_type
from neptune.types.atoms.datetime import Datetime as DatetimeVal

if typing.TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


class Datetime(CopiableAtom):
    @staticmethod
    def create_assignment_operation(path, value: datetime):
        return AssignDatetime(path, value)

    @staticmethod
    def getter(
        backend: "NeptuneBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
    ) -> datetime:
        val = backend.get_datetime_attribute(container_id, container_type, path)
        return val.value

    def assign(self, value: typing.Union[DatetimeVal, datetime, StringifyValue], *, wait: bool = False):
        verify_type("value", value, (DatetimeVal, datetime, StringifyValue))
        if isinstance(value, (DatetimeVal, StringifyValue)):
            value = value.value
        else:
            value = value.replace(microsecond=1000 * int(value.microsecond / 1000))
        with self._container.lock():
            self._enqueue_operation(self.create_assignment_operation(self._path, value), wait=wait)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Integer"]

import typing

from neptune.attributes.atoms.copiable_atom import CopiableAtom
from neptune.internal.container_type import ContainerType
from neptune.internal.operation import AssignInt
from neptune.internal.utils.logger import logger
from neptune.types.atoms.integer import Integer as IntegerVal

if typing.TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


class Integer(CopiableAtom):

    MAX_32_BIT_INT = 2147483647
    MIN_32_BIT_INT = -2147483648

    @staticmethod
    def create_assignment_operation(path, value: int):
        return AssignInt(path, value)

    @staticmethod
    def getter(
        backend: "NeptuneBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
    ) -> int:
        val = backend.get_int_attribute(container_id, container_type, path)
        return val.value

    def assign(self, value: typing.Union[IntegerVal, float, int], *, wait: bool = False):
        if not isinstance(value, IntegerVal):
            value = IntegerVal(value)

        if value.value < Integer.MIN_32_BIT_INT or value.value > Integer.MAX_32_BIT_INT:
            logger.warning(
                "WARNING: The value you're trying to log is outside the range of 32-bit integers "
                "(%s to %s) and will be skipped. "
                "We'll support 64-bit integers in the future. "
                'For now, try logging the value as a float instead: run["field"] = float(%s)',
                Integer.MIN_32_BIT_INT,
                Integer.MAX_32_BIT_INT,
                value.value,
            )
            return

        with self._container.lock():
            self._enqueue_operation(self.create_assignment_operation(self._path, value.value), wait=wait)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StringifyValue", "extract_if_stringify_value"]


from typing import Any


class StringifyValue:
    def __init__(self, value: Any):
        self.__value = value

    @property
    def value(self):
        return self.__value

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return repr(self.__value)


def extract_if_stringify_value(val):
    if isinstance(val, StringifyValue):
        return val.value
    return val



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Float"]

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

from neptune.internal.types.stringify_value import extract_if_stringify_value
from neptune.types.atoms.atom import Atom

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")


@dataclass
class Float(Atom):

    value: float

    def __init__(self, value):
        self.value = float(extract_if_stringify_value(value))

    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        return visitor.visit_float(self)

    def __str__(self):
        return "Float({})".format(str(self.value))

    def __float__(self):
        return self.value



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
import logging
import os
import stat
import time
from abc import (
    ABCMeta,
    abstractmethod,
)
from dataclasses import dataclass
from io import BytesIO
from pprint import pformat
from typing import (
    BinaryIO,
    Generator,
    List,
    Set,
    Union,
)

import six

_logger = logging.getLogger(__name__)


@dataclass
class AttributeUploadConfiguration:
    chunk_size: int


class UploadEntry(object):
    def __init__(self, source: Union[str, BytesIO], target_path: str):
        self.source = source
        self.target_path = target_path

    def length(self) -> int:
        if self.is_stream():
            return self.source.getbuffer().nbytes
        else:
            return os.path.getsize(self.source)

    def get_stream(self) -> Union[BinaryIO, io.BytesIO]:
        if self.is_stream():
            return self.source
        else:
            return io.open(self.source, "rb")

    def get_permissions(self) -> str:
        if self.is_stream():
            return "----------"
        else:
            return self.permissions_to_unix_string(self.source)

    @classmethod
    def permissions_to_unix_string(cls, path):
        st = 0
        if os.path.exists(path):
            st = os.lstat(path).st_mode
        is_dir = "d" if stat.S_ISDIR(st) else "-"
        dic = {
            "7": "rwx",
            "6": "rw-",
            "5": "r-x",
            "4": "r--",
            "3": "-wx",
            "2": "-w-",
            "1": "--x",
            "0": "---",
        }
        perm = ("%03o" % st)[-3:]
        return is_dir + "".join(dic.get(x, x) for x in perm)

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

    def __hash__(self):
        """
        Returns the hash of source and target path
        """
        return hash((self.source, self.target_path))

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.__dict__)

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def is_stream(self):
        return hasattr(self.source, "read")


class UploadPackage(object):
    def __init__(self):
        self.items: List[UploadEntry] = []
        self.size: int = 0
        self.len: int = 0

    def reset(self):
        self.items = []
        self.size = 0
        self.len = 0

    def update(self, entry: UploadEntry, size: int):
        self.items.append(entry)
        self.size += size
        self.len += 1

    def is_empty(self):
        return self.len == 0

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.__dict__)

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()


@six.add_metaclass(ABCMeta)
class ProgressIndicator(object):
    @abstractmethod
    def progress(self, steps):
        pass

    @abstractmethod
    def complete(self):
        pass


class LoggingProgressIndicator(ProgressIndicator):
    def __init__(self, total, frequency=10):
        self.current = 0
        self.total = total
        self.last_warning = time.time()
        self.frequency = frequency
        _logger.warning(
            "You are sending %dMB of source code to Neptune. "
            "It is pretty uncommon - please make sure it's what you wanted.",
            self.total / (1024 * 1024),
        )

    def progress(self, steps):
        self.current += steps
        if time.time() - self.last_warning > self.frequency:
            _logger.warning(
                "%d MB / %d MB (%d%%) of source code was sent to Neptune.",
                self.current / (1024 * 1024),
                self.total / (1024 * 1024),
                100 * self.current / self.total,
            )
            self.last_warning = time.time()

    def complete(self):
        _logger.warning(
            "%d MB (100%%) of source code was sent to Neptune.",
            self.total / (1024 * 1024),
        )


class SilentProgressIndicator(ProgressIndicator):
    def __init__(self):
        pass

    def progress(self, steps):
        pass

    def complete(self):
        pass


def scan_unique_upload_entries(upload_entries):
    """
    Returns upload entries for all files that could be found for given upload entries.
    In case of directory as upload entry, files we be taken from all subdirectories recursively.
    Any duplicated entries are removed.
    """
    walked_entries = set()
    for entry in upload_entries:
        if entry.is_stream() or not os.path.isdir(entry.source):
            walked_entries.add(entry)
        else:
            for root, _, files in os.walk(entry.source):
                path_relative_to_entry_source = os.path.relpath(root, entry.source)
                target_root = os.path.normpath(os.path.join(entry.target_path, path_relative_to_entry_source))
                for filename in files:
                    walked_entries.add(
                        UploadEntry(
                            os.path.join(root, filename),
                            os.path.join(target_root, filename),
                        )
                    )

    return walked_entries


def split_upload_files(
    upload_entries: Set[UploadEntry],
    upload_configuration: AttributeUploadConfiguration,
    max_files=500,
) -> Generator[UploadPackage, None, None]:
    current_package = UploadPackage()

    for entry in upload_entries:
        if entry.is_stream():
            if current_package.len > 0:
                yield current_package
                current_package.reset()
            current_package.update(entry, 0)
            yield current_package
            current_package.reset()
        else:
            size = os.path.getsize(entry.source)
            if (
                size + current_package.size > upload_configuration.chunk_size or current_package.len > max_files
            ) and not current_package.is_empty():
                yield current_package
                current_package.reset()
            current_package.update(entry, size)

    yield current_package


def normalize_file_name(name):
    return name.replace(os.sep, "/")



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Log and organize all your ML model metadata with neptune.ai.

There are four kinds of Neptune object: run, model, model version, and project.
These help you track, store, and visualize metadata related to your model-training experiments.
The package contains the functions and constructors needed to initialize the objects.
You can either create new objects or connect to existing ones (to, for example, fetch or add more metadata).

Functions:
    init_run()
    init_model()
    init_model_version()
    init_project()

Classes:
    Run
    Model
    ModelVersion
    Project

Constants:
    ANONYMOUS_API_TOKEN

Tracking runs
-------------
A Neptune run tracks some things automatically during the execution of your model training
script, such as hardware consumption, source code, and Git information. You can also
assign any metadata to the run manually and organize it in a structure of your choosing.

>>> run = neptune.init_run()
>>> run["some/structure"] = some_metadata

Model registry
--------------
Create a model object to register a model:

>>> model = neptune.init_model(key="MOD")
>>> model["signature"].upload("signature.json")

Then create as many versions of the model as you need, tracking their metadata and
lifecycles separately:

>>> model_version = neptune.init_model_version(model="PROJ-MOD")
>>> model_version["dataset_version"].track_files("./data/train.csv")
>>> model_version.change_stage("staging")

Project metadata
----------------
Initialize your entire Neptune project and log metadata on project-level:

>>> project = neptune.init_project(project="ml-team/classification")
>>> project["datasets"].upload("./data/")

Initializing with class constructor
-----------------------------------
You can also use the class constructor to initialize a Neptune object.

>>> from neptune import Run
>>> run = Run()

>>> from neptune import ModelVersion
>>> model_version = ModelVersion(with_id="PROJ-MOD-3")  # connect to existing model version
>>> model_version.change_stage("production")

Anonymous logging
-----------------
To try out Neptune without registering, you can pass the `ANONYMOUS_API_TOKEN` constant
to the `api_token` argument when initializing Neptune.

>>> with neptune.init_run(api_token=neptune.ANONYMOUS_API_TOKEN) as run:
...     ...

---

Learn more in the docs: https://docs.neptune.ai/api/neptune/
"""
__all__ = [
    "ANONYMOUS_API_TOKEN",
    "init_model",
    "init_model_version",
    "init_project",
    "init_run",
    "Run",
    "Model",
    "ModelVersion",
    "Project",
    "__version__",
]


from neptune.common.patches import apply_patches
from neptune.constants import ANONYMOUS_API_TOKEN
from neptune.metadata_containers import (
    Model,
    ModelVersion,
    Project,
    Run,
)
from neptune.version import __version__

# Apply patches of external libraries
apply_patches()

init_run = Run
init_model = Model
init_model_version = ModelVersion
init_project = Project



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["GPUMonitor"]

import logging

from neptune.vendor.pynvml import (
    NVMLError,
    nvmlDeviceGetCount,
    nvmlDeviceGetHandleByIndex,
    nvmlDeviceGetMemoryInfo,
    nvmlDeviceGetUtilizationRates,
    nvmlInit,
)

_logger = logging.getLogger(__name__)


class GPUMonitor(object):

    nvml_error_printed = False

    def get_card_count(self):
        return self.__nvml_get_or_else(nvmlDeviceGetCount, default=0)

    def get_card_usage_percent(self, card_index):
        return self.__nvml_get_or_else(
            lambda: float(nvmlDeviceGetUtilizationRates(nvmlDeviceGetHandleByIndex(card_index)).gpu)
        )

    def get_card_used_memory_in_bytes(self, card_index):
        return self.__nvml_get_or_else(lambda: nvmlDeviceGetMemoryInfo(nvmlDeviceGetHandleByIndex(card_index)).used)

    def get_top_card_memory_in_bytes(self):
        def read_top_card_memory_in_bytes():
            return self.__nvml_get_or_else(
                lambda: [
                    nvmlDeviceGetMemoryInfo(nvmlDeviceGetHandleByIndex(card_index)).total
                    for card_index in range(nvmlDeviceGetCount())
                ],
                default=0,
            )

        memory_per_card = read_top_card_memory_in_bytes()
        if not memory_per_card:
            return 0
        return max(memory_per_card)

    def __nvml_get_or_else(self, getter, default=None):
        try:
            nvmlInit()
            return getter()
        except NVMLError as e:
            if not GPUMonitor.nvml_error_printed:
                warning = (
                    "Info (NVML): %s. GPU usage metrics may not be reported. For more information, "
                    "see https://docs.neptune.ai/help/nvml_error/"
                )
                _logger.warning(warning, e)
                GPUMonitor.nvml_error_printed = True
            return default



__all__ = [
    "DependencyTrackingStrategy",
    "InferDependenciesStrategy",
    "FileDependenciesStrategy",
]

import os
import sys
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Union,
)

if sys.version_info >= (3, 8):
    from importlib.metadata import distributions
else:
    from importlib_metadata import distributions

from neptune.internal.utils.logger import logger
from neptune.types import File

if TYPE_CHECKING:
    from neptune import Run


class DependencyTrackingStrategy(ABC):
    @abstractmethod
    def log_dependencies(self, run: "Run") -> None:
        ...


class InferDependenciesStrategy(DependencyTrackingStrategy):
    def log_dependencies(self, run: "Run") -> None:
        dependencies = []
        dists = list(sorted(distributions(), key=lambda d: d.metadata["Name"]))

        for dist in dists:
            name, version = dist.metadata["Name"], dist.metadata["Version"]
            dependencies.append(f"{name}=={version}")

        dependencies_str = "\n".join(dependencies)

        if dependencies_str:
            run["source_code/requirements"].upload(File.from_content(dependencies_str))


class FileDependenciesStrategy(DependencyTrackingStrategy):
    def __init__(self, path: Union[str, os.PathLike]):
        self._path = path

    def log_dependencies(self, run: "Run") -> None:
        if os.path.isfile(self._path):
            run["source_code/requirements"].upload(os.path.basename(self._path))
        else:
            logger.error("[ERROR] File '%s' does not exist - skipping dependency file upload.", self._path)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from dataclasses import dataclass
from enum import Enum

from neptune.internal.utils import verify_type
from neptune.management.exceptions import UnsupportedValue
from neptune.management.internal.types import (
    ProjectMemberRole,
    ProjectVisibility,
    WorkspaceMemberRole,
)


class ProjectVisibilityDTO(Enum):
    PRIVATE = "priv"
    PUBLIC = "pub"
    WORKSPACE = "workspace"

    @classmethod
    def from_str(cls, visibility: str) -> "ProjectVisibilityDTO":
        verify_type("visibility", visibility, str)

        try:
            return {
                ProjectVisibility.PRIVATE: ProjectVisibilityDTO.PRIVATE,
                ProjectVisibility.PUBLIC: ProjectVisibilityDTO.PUBLIC,
                ProjectVisibility.WORKSPACE: ProjectVisibilityDTO.WORKSPACE,
            }[visibility]
        except KeyError as e:
            raise UnsupportedValue(enum=cls.__name__, value=visibility) from e


class ProjectMemberRoleDTO(Enum):
    VIEWER = "viewer"
    MEMBER = "member"
    MANAGER = "manager"

    @classmethod
    def from_str(cls, role: str) -> "ProjectMemberRoleDTO":
        verify_type("role", role, str)

        try:
            return {
                ProjectMemberRole.VIEWER: ProjectMemberRoleDTO.VIEWER,
                ProjectMemberRole.CONTRIBUTOR: ProjectMemberRoleDTO.MEMBER,
                ProjectMemberRole.OWNER: ProjectMemberRoleDTO.MANAGER,
            }[role]
        except KeyError as e:
            raise UnsupportedValue(enum=cls.__name__, value=role) from e

    @staticmethod
    def to_domain(role: str) -> str:
        verify_type("role", role, str)

        return {
            ProjectMemberRoleDTO.VIEWER.value: ProjectMemberRole.VIEWER,
            ProjectMemberRoleDTO.MANAGER.value: ProjectMemberRole.OWNER,
            ProjectMemberRoleDTO.MEMBER.value: ProjectMemberRole.CONTRIBUTOR,
        }.get(role)


class WorkspaceMemberRoleDTO(Enum):
    OWNER = "owner"
    MEMBER = "member"

    @staticmethod
    def to_domain(role: str) -> str:
        return {
            WorkspaceMemberRoleDTO.OWNER.value: WorkspaceMemberRole.ADMIN,
            WorkspaceMemberRoleDTO.MEMBER.value: WorkspaceMemberRole.MEMBER,
        }.get(role)


@dataclass
class ServiceAccountDTO:
    name: str
    id: str



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from collections import namedtuple

from neptune.attributes import constants as alpha_consts
from neptune.internal import operation as alpha_operation
from neptune.internal.backends.api_model import AttributeType as AlphaAttributeType

# Alpha equivalent of old api's `KeyValueProperty` used in `Experiment.properties`
from neptune.internal.operation import ImageValue
from neptune.legacy.exceptions import NeptuneException
from neptune.legacy.internal.channels.channels import (
    ChannelType,
    ChannelValueType,
)

AlphaKeyValueProperty = namedtuple("AlphaKeyValueProperty", ["key", "value"])


class AlphaAttributeWrapper(abc.ABC):
    """It's simple wrapper for `AttributeDTO`."""

    _allowed_atribute_types = list()

    def __init__(self, attribute):
        """Expects `AttributeDTO`"""
        assert self._allowed_atribute_types is not None
        if not self.is_valid_attribute(attribute):
            raise NeptuneException(f"Invalid channel attribute type: {attribute.type}")

        self._attribute = attribute

    @classmethod
    def is_valid_attribute(cls, attribute):
        """Checks if attribute can be wrapped by particular descendant of this class"""
        return attribute.type in cls._allowed_atribute_types

    @property
    def _properties(self):
        """Returns proper attribute property according to type"""
        return getattr(self._attribute, f"{self._attribute.type}Properties")


class AlphaPropertyDTO(AlphaAttributeWrapper):
    """It's simple wrapper for `AttributeDTO` objects which uses alpha variables attributes to fake properties.

    Alpha leaderboard doesn't have `KeyValueProperty` since it doesn't support properties at all,
    so we do need fake `KeyValueProperty` class for backward compatibility with old client's code."""

    _allowed_atribute_types = [
        AlphaAttributeType.STRING.value,
    ]

    @classmethod
    def is_valid_attribute(cls, attribute):
        """Checks if attribute can be used as property"""
        has_valid_type = super().is_valid_attribute(attribute)
        is_in_properties_space = attribute.name.startswith(alpha_consts.PROPERTIES_ATTRIBUTE_SPACE)
        return has_valid_type and is_in_properties_space

    @property
    def key(self):
        return self._properties.attributeName.split("/", 1)[-1]

    @property
    def value(self):
        return self._properties.value


class AlphaParameterDTO(AlphaAttributeWrapper):
    """It's simple wrapper for `AttributeDTO` objects which uses alpha variables attributes to fake properties.

    Alpha leaderboard doesn't have `KeyValueProperty` since it doesn't support properties at all,
    so we do need fake `KeyValueProperty` class for backward compatibility with old client's code."""

    _allowed_atribute_types = [
        AlphaAttributeType.FLOAT.value,
        AlphaAttributeType.STRING.value,
        AlphaAttributeType.DATETIME.value,
    ]

    @classmethod
    def is_valid_attribute(cls, attribute):
        """Checks if attribute can be used as property"""
        has_valid_type = super().is_valid_attribute(attribute)
        is_in_parameters_space = attribute.name.startswith(alpha_consts.PARAMETERS_ATTRIBUTE_SPACE)
        return has_valid_type and is_in_parameters_space

    @property
    def name(self):
        return self._properties.attributeName.split("/", 1)[-1]

    @property
    def value(self):
        return self._properties.value

    @property
    def parameterType(self):
        return "double" if self._properties.attributeType == AlphaAttributeType.FLOAT.value else "string"


class AlphaChannelDTO(AlphaAttributeWrapper):
    """It's simple wrapper for `AttributeDTO` objects which uses alpha series attributes to fake channels.

    Alpha leaderboard doesn't have `ChannelDTO` since it doesn't support channels at all,
    so we do need fake `ChannelDTO` class for backward compatibility with old client's code."""

    _allowed_atribute_types = [
        AlphaAttributeType.FLOAT_SERIES.value,
        AlphaAttributeType.STRING_SERIES.value,
        AlphaAttributeType.IMAGE_SERIES.value,
    ]

    @property
    def id(self):
        return self._properties.attributeName

    @property
    def name(self):
        return self._properties.attributeName.split("/", 1)[-1]

    @property
    def channelType(self):
        attr_type = self._properties.attributeType
        if attr_type == AlphaAttributeType.FLOAT_SERIES.value:
            return ChannelType.NUMERIC.value
        elif attr_type == AlphaAttributeType.STRING_SERIES.value:
            return ChannelType.TEXT.value
        elif attr_type == AlphaAttributeType.IMAGE_SERIES.value:
            return ChannelType.IMAGE.value

    @property
    def x(self):
        return self._properties.lastStep

    @property
    def y(self):
        if self.channelType == ChannelType.IMAGE.value:
            # We do not store last value for image series
            return None
        return self._properties.last


class AlphaChannelWithValueDTO:
    """Alpha leaderboard doesn't have `ChannelWithValueDTO` since it doesn't support channels at all,
    so we do need fake `ChannelWithValueDTO` class for backward compatibility with old client's code"""

    def __init__(self, channelId: str, channelName: str, channelType: str, x, y):
        self._ch_id = channelId
        self._ch_name = channelName
        self._ch_type = channelType
        self._x = x
        self._y = y

    @property
    def channelId(self):
        return self._ch_id

    @property
    def channelName(self):
        return self._ch_name

    @property
    def channelType(self):
        return self._ch_type

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = x

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        self._y = y


def _map_using_dict(el, el_name, source_dict) -> alpha_operation.Operation:
    try:
        return source_dict[el]
    except KeyError as e:
        raise NeptuneException(f"We're not supporting {el} {el_name}.") from e


def channel_type_to_operation(channel_type: ChannelType) -> alpha_operation.Operation:
    _channel_type_to_operation = {
        ChannelType.TEXT: alpha_operation.LogStrings,
        ChannelType.NUMERIC: alpha_operation.LogFloats,
        ChannelType.IMAGE: alpha_operation.LogImages,
    }
    return _map_using_dict(channel_type, "channel type", _channel_type_to_operation)


def channel_type_to_clear_operation(
    channel_type: ChannelType,
) -> alpha_operation.Operation:
    _channel_type_to_operation = {
        ChannelType.TEXT: alpha_operation.ClearStringLog,
        ChannelType.NUMERIC: alpha_operation.ClearFloatLog,
        ChannelType.IMAGE: alpha_operation.ClearImageLog,
    }
    return _map_using_dict(channel_type, "channel type", _channel_type_to_operation)


def channel_value_type_to_operation(
    channel_value_type: ChannelValueType,
) -> alpha_operation.Operation:
    _channel_value_type_to_operation = {
        ChannelValueType.TEXT_VALUE: alpha_operation.LogStrings,
        ChannelValueType.NUMERIC_VALUE: alpha_operation.LogFloats,
        ChannelValueType.IMAGE_VALUE: alpha_operation.LogImages,
    }
    return _map_using_dict(channel_value_type, "channel value type", _channel_value_type_to_operation)


def deprecated_img_to_alpha_image(img: dict) -> ImageValue:
    return ImageValue(data=img["data"], name=img["name"], description=img["description"])



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "NQLQuery",
    "NQLAggregator",
    "NQLQueryAggregate",
    "NQLAttributeOperator",
    "NQLAttributeType",
    "NQLQueryAttribute",
]

import typing
from dataclasses import dataclass
from enum import Enum
from typing import Iterable


@dataclass
class NQLQuery:
    pass


class NQLEmptyQuery(NQLQuery):
    def __str__(self) -> str:
        return ""


class NQLAggregator(str, Enum):
    AND = "AND"
    OR = "OR"


@dataclass
class NQLQueryAggregate(NQLQuery):
    items: Iterable[NQLQuery]
    aggregator: NQLAggregator

    def __str__(self) -> str:
        if self.items:
            return "(" + f" {self.aggregator.value} ".join(map(str, self.items)) + ")"
        return ""


class NQLAttributeOperator(str, Enum):
    EQUALS = "="
    CONTAINS = "CONTAINS"


class NQLAttributeType(str, Enum):
    STRING = "string"
    STRING_SET = "stringSet"
    EXPERIMENT_STATE = "experimentState"
    BOOLEAN = "bool"


@dataclass
class NQLQueryAttribute(NQLQuery):
    name: str
    type: NQLAttributeType
    operator: NQLAttributeOperator
    value: typing.Union[str, bool]

    def __str__(self) -> str:
        if isinstance(self.value, bool):
            value = str(self.value).lower()
        else:
            value = f'"{self.value}"'

        return f"(`{self.name}`:{self.type.value} {self.operator.value} {value})"



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import os.path
import sys
from typing import (
    List,
    Optional,
    Tuple,
)

from neptune.common.storage.storage_utils import normalize_file_name
from neptune.common.utils import (
    glob,
    is_ipython,
)


def get_source_code_to_upload(
    upload_source_files: Optional[List[str]],
) -> Tuple[str, List[Tuple[str, str]]]:
    source_target_pairs = []
    if is_ipython():
        main_file = None
        entrypoint = None
    else:
        main_file = sys.argv[0]
        entrypoint = main_file or None
    if upload_source_files is None:
        if main_file is not None and os.path.isfile(main_file):
            entrypoint = normalize_file_name(os.path.basename(main_file))
            source_target_pairs = [
                (
                    os.path.abspath(main_file),
                    normalize_file_name(os.path.basename(main_file)),
                )
            ]
    else:
        expanded_source_files = set()
        for filepath in upload_source_files:
            expanded_source_files |= set(glob(filepath))
        if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 5):
            for filepath in expanded_source_files:
                if filepath.startswith(".."):
                    raise ValueError("You need to have Python 3.5 or later to use paths outside current directory.")
                source_target_pairs.append((os.path.abspath(filepath), normalize_file_name(filepath)))
        else:
            absolute_paths = []
            for filepath in expanded_source_files:
                absolute_paths.append(os.path.abspath(filepath))
            try:
                common_source_root = os.path.commonpath(absolute_paths)
            except ValueError:
                for absolute_path in absolute_paths:
                    source_target_pairs.append((absolute_path, normalize_file_name(absolute_path)))
            else:
                if os.path.isfile(common_source_root):
                    common_source_root = os.path.dirname(common_source_root)
                if common_source_root.startswith(os.getcwd() + os.sep):
                    common_source_root = os.getcwd()
                for absolute_path in absolute_paths:
                    source_target_pairs.append(
                        (
                            absolute_path,
                            normalize_file_name(os.path.relpath(absolute_path, common_source_root)),
                        )
                    )
    return entrypoint, source_target_pairs



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["cast_value", "cast_value_for_extend"]

import argparse
from datetime import datetime
from typing import (
    Any,
    Collection,
    Optional,
    Union,
)

from neptune.internal.types.stringify_value import StringifyValue
from neptune.internal.utils import (
    is_bool,
    is_dict_like,
    is_float,
    is_float_like,
    is_int,
    is_string,
    is_stringify_value,
)
from neptune.types import (
    Boolean,
    File,
    Integer,
)
from neptune.types.atoms.datetime import Datetime
from neptune.types.atoms.float import Float
from neptune.types.atoms.string import String
from neptune.types.namespace import Namespace
from neptune.types.series import (
    FileSeries,
    FloatSeries,
    StringSeries,
)
from neptune.types.series.series import Series
from neptune.types.value import Value
from neptune.types.value_copy import ValueCopy


def cast_value(value: Any) -> Optional[Value]:
    from neptune.handler import Handler

    from_stringify_value = False
    if is_stringify_value(value):
        from_stringify_value, value = True, value.value

    if isinstance(value, Value):
        return value
    elif isinstance(value, Handler):
        return ValueCopy(value)
    elif isinstance(value, argparse.Namespace):
        return Namespace(vars(value))
    elif File.is_convertable_to_image(value):
        return File.as_image(value)
    elif File.is_convertable_to_html(value):
        return File.as_html(value)
    elif is_bool(value):
        return Boolean(value)
    elif is_int(value):
        return Integer(value)
    elif is_float(value):
        return Float(value)
    elif is_string(value):
        return String(value)
    elif isinstance(value, datetime):
        return Datetime(value)
    elif is_float_like(value):
        return Float(value)
    elif is_dict_like(value):
        return Namespace(value)
    elif from_stringify_value:
        return String(str(value))


def cast_value_for_extend(
    values: Union[StringifyValue, Namespace, Series, Collection[Any]]
) -> Optional[Union[Series, Namespace]]:
    from_stringify_value, original_values = False, None
    if is_stringify_value(values):
        from_stringify_value, original_values, values = True, values, values.value

    if isinstance(values, Namespace):
        return values
    elif is_dict_like(values):
        return Namespace(values)
    elif isinstance(values, Series):
        return values

    sample_val = next(iter(values))

    if isinstance(sample_val, File):
        return FileSeries(values=values)
    elif File.is_convertable_to_image(sample_val):
        return FileSeries(values=values)
    elif File.is_convertable_to_html(sample_val):
        return FileSeries(values=values)
    elif is_string(sample_val):
        return StringSeries(values=values)
    elif is_float_like(sample_val):
        return FloatSeries(values=values)
    elif from_stringify_value:
        return StringSeries(values=original_values)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools
import re

import bravado_core.model
from bravado_core.model import (
    _bless_models,
    _collect_models,
    _get_unprocessed_uri,
    _post_process_spec,
    _tag_models,
)


def _run_post_processing(spec):
    visited_models = {}

    def _call_post_process_spec(spec_dict):
        # Discover all the models in spec_dict
        _post_process_spec(
            spec_dict=spec_dict,
            spec_resolver=spec.resolver,
            on_container_callbacks=[
                functools.partial(
                    _tag_models,
                    visited_models=visited_models,
                    swagger_spec=spec,
                ),
                functools.partial(
                    _bless_models,
                    visited_models=visited_models,
                    swagger_spec=spec,
                ),
                functools.partial(
                    _collect_models,
                    models=spec.definitions,
                    swagger_spec=spec,
                ),
            ],
        )

    # Post process specs to identify models
    _call_post_process_spec(spec.spec_dict)

    processed_uris = {
        uri
        for uri in spec.resolver.store
        if uri == spec.origin_url or re.match(r"http(s)?://json-schema\.org/draft(/\d{4})?-\d+/(schema|meta/.*)", uri)
    }
    additional_uri = _get_unprocessed_uri(spec, processed_uris)
    while additional_uri is not None:
        # Post process each referenced specs to identify models in definitions of linked files
        with spec.resolver.in_scope(additional_uri):
            _call_post_process_spec(
                spec.resolver.store[additional_uri],
            )

        processed_uris.add(additional_uri)
        additional_uri = _get_unprocessed_uri(spec, processed_uris)


# Issue: https://github.com/Yelp/bravado-core/issues/388
# Bravado currently makes additional requests to `json-schema.org` in order to gather mission schemas
# This makes `neptune` unable to run without internet connection or with a many security policies
def patch():
    bravado_core.model._run_post_processing = _run_post_processing



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from neptune.legacy.internal.channels.channels import ChannelNamespace
from neptune.legacy.internal.streams.channel_writer import ChannelWriter


class StdStreamWithUpload(object):
    def __init__(self, experiment, channel_name, stream):
        self._channel = experiment._get_channel(channel_name, "text", ChannelNamespace.SYSTEM)
        self._channel_writer = ChannelWriter(experiment, channel_name, ChannelNamespace.SYSTEM)
        self._stream = stream

    def write(self, data):
        self._stream.write(data)
        try:
            self._channel_writer.write(data)
        except:  # noqa: E722
            pass

    def isatty(self):
        return hasattr(self._stream, "isatty") and self._stream.isatty()

    def flush(self):
        self._stream.flush()

    def fileno(self):
        return self._stream.fileno()


class StdOutWithUpload(StdStreamWithUpload):
    def __init__(self, experiment):
        super(StdOutWithUpload, self).__init__(experiment, "stdout", sys.__stdout__)
        sys.stdout = self

    def close(self):
        sys.stdout = sys.__stdout__


class StdErrWithUpload(StdStreamWithUpload):
    def __init__(self, experiment):
        super(StdErrWithUpload, self).__init__(experiment, "stderr", sys.__stderr__)
        sys.stderr = self

    def close(self):
        sys.stderr = sys.__stderr__



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

REGISTERED_CODES = dict()


class ManagementOperationFailure(Exception):
    code = -1
    description = "Unknown error"

    def __init__(self, **kwargs):
        super().__init__()
        self._properties: dict = kwargs or {}

    def __str__(self):
        return f"{self.description.format(**self._properties)} (code: {self.code})"

    def __init_subclass__(cls):
        previous = REGISTERED_CODES.get(cls.code)
        assert previous is None, f"{cls} cannot have code {cls.code} already used by {previous}"
        REGISTERED_CODES[cls.code] = cls

    @property
    def details(self):
        return {
            "code": self.code,
            "description": self.description.format(**self._properties),
        }


class InvalidProjectName(ManagementOperationFailure):
    code = 1
    description = 'Provided project name "{name}" could not be parsed.'


class MissingWorkspaceName(ManagementOperationFailure):
    code = 2
    description = 'Cannot resolve project "{name}", you have to provide a workspace name.'


class ConflictingWorkspaceName(ManagementOperationFailure):
    code = 3
    description = 'Project name "{name}" conflicts with provided workspace "{workspace}".'


class ProjectNotFound(ManagementOperationFailure):
    code = 4
    description = 'Project "{name}" could not be found.'


class WorkspaceNotFound(ManagementOperationFailure):
    code = 5
    description = 'Workspace "{workspace}" could not be found.'


class ProjectAlreadyExists(ManagementOperationFailure):
    code = 6
    description = 'Project "{name}" already exists.'


class AccessRevokedOnDeletion(ManagementOperationFailure):
    code = 7
    description = 'Not enough permissions to delete project "{name}".'


class AccessRevokedOnMemberRemoval(ManagementOperationFailure):
    code = 8
    description = 'Not enough permissions to remove user "{user}" from project "{project}".'


class UserNotExistsOrWithoutAccess(ManagementOperationFailure):
    code = 9
    description = (
        'User "{user}" does not exist or has no access to project "{project}". '
        "If the project visibility is set to 'workspace', a user cannot be added or removed."
    )


class UserAlreadyHasAccess(ManagementOperationFailure):
    code = 10
    description = 'User "{user}" already has access to the project "{project}". Their role is "{role}".'


class ProjectsLimitReached(ManagementOperationFailure):
    code = 11
    description = "Project number limit reached."


class UnsupportedValue(ManagementOperationFailure):
    code = 12
    description = "{enum} cannot have value {value}"


class ServiceAccountAlreadyHasAccess(ManagementOperationFailure):
    code = 13
    description = (
        'The service account "{service_account_name}" already has access to the project "{project}", '
        "either because it was already added or because of the project's visibility setting. "
        'The role of the service account is "{role}".'
    )


class AccessRevokedOnServiceAccountRemoval(ManagementOperationFailure):
    code = 14
    description = (
        'Not enough permissions to remove service account "{service_account_name}" from project "{project}". '
        "The account that performs the removal must be a project owner."
    )


class ServiceAccountNotExistsOrWithoutAccess(ManagementOperationFailure):
    code = 15
    description = (
        'Service account "{service_account_name}" does not exist or did not have access to project "{project}" '
        'in the first place. If the project visibility is set to "workspace", you cannot revoke access for '
        "invididual workspace members."
    )


class ServiceAccountNotFound(ManagementOperationFailure):
    code = 16
    description = 'Service account "{service_account_name}" could not be found in workspace "{workspace}"'


class ProjectKeyCollision(ManagementOperationFailure):
    code = 17
    description = 'Project with key "{key}" already exists.'


class ProjectNameCollision(ManagementOperationFailure):
    code = 18
    description = 'Project with name "{name}" already exists.'


class ProjectKeyInvalid(ManagementOperationFailure):
    code = 19
    description = 'Invalid project key "{key}": {reason}'


class ProjectNameInvalid(ManagementOperationFailure):
    code = 20
    description = 'Invalid project name "{name}": {reason}'


class BadRequestException(ManagementOperationFailure):
    code = 400
    description = "Your request has encountered the following validation errors: {validation_errors}"


class IncorrectIdentifierException(ManagementOperationFailure):
    code = 21
    description = "Can not parse '{identifier}' as identifier."


class ObjectNotFound(ManagementOperationFailure):
    code = 22
    description = "Object not found."


class WorkspaceOrUserNotFound(ManagementOperationFailure):
    code = 23
    description = "Workspace '{workspace}' or user '{user}' could not be found."


class UserAlreadyInvited(ManagementOperationFailure):
    code = 24
    description = "User '{user}' has already been invited to the workspace '{workspace}'."


class ProjectPrivacyRestrictedException(ManagementOperationFailure):
    code = 25
    description = (
        "Cannot set {requested} visibility for project. {followup}This might be caused by workspace "
        "settings or limited by your plan."
    )

    def __init__(self, **kwargs):
        modified_kwargs = {"followup": ""}
        allowed = kwargs.get("allowed")
        if allowed and isinstance(allowed, list):
            modified_kwargs["followup"] = "Allowed values are: {allowed}. ".format(
                allowed=", ".join(['"' + option + '"' for option in allowed])
            )
        modified_kwargs.update(kwargs)
        requested = modified_kwargs.get("requested")
        if not requested:
            modified_kwargs["requested"] = "the selected"
        else:
            modified_kwargs["requested"] = '"' + requested + '"'
        super().__init__(**modified_kwargs)


class ActiveProjectsLimitReachedException(ManagementOperationFailure):
    code = 26
    description = (
        "Limit of active projects reached. You can have up to {currentQuota} active projects simultaneously. "
        "To create a new project, you need to either archive an active project or increase the quota of active "
        "projects in the workspace."
    )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["expect_not_an_experiment", "join_paths", "verify_type", "RunType"]

from typing import Union

from neptune import Run
from neptune.common.experiments import LegacyExperiment as Experiment
from neptune.exceptions import NeptuneLegacyIncompatibilityException
from neptune.handler import Handler
from neptune.internal.utils import verify_type
from neptune.internal.utils.paths import join_paths


def expect_not_an_experiment(run: Run):
    if isinstance(run, Experiment):
        raise NeptuneLegacyIncompatibilityException()


RunType = Union[Run, Handler]



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["StatusRunner"]

import sys
import textwrap
from pathlib import Path
from typing import (
    Optional,
    Sequence,
)

from neptune.cli.abstract_backend_runner import AbstractBackendRunner
from neptune.cli.container_manager import ContainersManager
from neptune.cli.utils import (
    get_offline_dirs,
    get_qualified_name,
)
from neptune.constants import OFFLINE_NAME_PREFIX
from neptune.envs import PROJECT_ENV_NAME
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.utils.logger import logger

offline_run_explainer = """
Runs which execute offline are not created on the server and they are not assigned to projects;
instead, they are identified by UUIDs like the ones above.
When synchronizing offline runs, please specify the workspace and project using the "--project"
flag. Alternatively, you can set the environment variable
{} to the target workspace/project. See the examples below.
""".format(
    PROJECT_ENV_NAME
)


class StatusRunner(AbstractBackendRunner):
    @staticmethod
    def list_containers(
        base_path: Path,
        synced_containers: Optional[Sequence[ApiExperiment]],
        unsynced_containers: Sequence[ApiExperiment],
        offline_dirs: Sequence[str],
    ) -> None:

        if not synced_containers and not unsynced_containers and not offline_dirs:
            logger.info("There are no Neptune objects in %s", base_path)
            sys.exit(1)

        StatusRunner.log_unsync_objects(unsynced_containers)

        StatusRunner.log_offline_objects(offline_dirs)

        if not unsynced_containers:
            logger.info("\nThere are no unsynchronized objects in %s", base_path)

        if not synced_containers:
            logger.info("\nThere are no synchronized objects in %s", base_path)

        logger.info("\nPlease run with the `neptune sync --help` to see example commands.")

    @staticmethod
    def trashed(cont: ApiExperiment):
        return " (Trashed)" if cont.trashed else ""

    @staticmethod
    def log_offline_objects(offline_dirs, info=True):
        if offline_dirs:
            logger.info("Unsynchronized offline objects:")
            for container_id in offline_dirs:
                logger.info("- %s", f"{OFFLINE_NAME_PREFIX}{container_id}")
            if info:
                logger.info("\n%s", textwrap.fill(offline_run_explainer, width=90))

    @staticmethod
    def log_unsync_objects(unsynced_containers):
        if unsynced_containers:
            logger.info("Unsynchronized objects:")
            for container in unsynced_containers:
                logger.info("- %s%s", get_qualified_name(container), StatusRunner.trashed(container))

    def synchronization_status(self, base_path: Path) -> None:
        container_manager = ContainersManager(self._backend, base_path)
        synced_containers, unsynced_containers, not_found = container_manager.partition_containers_and_clean_junk(
            base_path
        )
        if len(not_found) > 0:
            logger.warning(
                "\nWARNING: %s objects was skipped because they do not exist anymore.",
                len(not_found),
            )
        offline_dirs = get_offline_dirs(base_path)
        self.list_containers(base_path, synced_containers, unsynced_containers, offline_dirs)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Artifact"]

import pathlib
import typing

from neptune.attributes.atoms.atom import Atom
from neptune.internal.artifacts.types import (
    ArtifactDriver,
    ArtifactDriversMap,
    ArtifactFileData,
)
from neptune.internal.backends.api_model import OptionalFeatures
from neptune.internal.operation import (
    AssignArtifact,
    TrackFilesToArtifact,
)
from neptune.types.atoms.artifact import Artifact as ArtifactVal


class Artifact(Atom):
    def _check_feature(self):
        self._container._backend.verify_feature_available(OptionalFeatures.ARTIFACTS)

    def assign(self, value: ArtifactVal, *, wait: bool = False):
        self._check_feature()
        # this function should be used only with ArtifactVal
        if not isinstance(value, ArtifactVal):
            raise TypeError("Value of unsupported type {}".format(type(value)))

        with self._container.lock():
            self._enqueue_operation(AssignArtifact(self._path, value.hash), wait=wait)

    def fetch(self) -> ArtifactVal:
        self._check_feature()
        return ArtifactVal(self.fetch_hash())

    def fetch_hash(self) -> str:
        self._check_feature()
        val = self._backend.get_artifact_attribute(self._container_id, self._container_type, self._path)
        return val.hash

    def fetch_files_list(self) -> typing.List[ArtifactFileData]:
        self._check_feature()
        artifact_hash = self.fetch_hash()
        return self._backend.list_artifact_files(
            self._container._project_id,
            artifact_hash,
        )

    def download(self, destination: str = None):
        self._check_feature()
        for file_definition in self.fetch_files_list():
            driver: typing.Type[ArtifactDriver] = ArtifactDriversMap.match_type(file_definition.type)
            file_destination = pathlib.Path(destination or ".") / pathlib.Path(file_definition.file_path)
            file_destination.parent.mkdir(parents=True, exist_ok=True)
            driver.download_file(file_destination, file_definition)

    def track_files(self, path: str, *, destination: str = None, wait: bool = False):
        self._check_feature()
        with self._container.lock():
            self._enqueue_operation(
                TrackFilesToArtifact(
                    self._path,
                    self._container._project_id,
                    [(path, destination)],
                ),
                wait=wait,
            )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from functools import wraps

from neptune.common.warnings import warn_once


def legacy_client_deprecation(func):
    @wraps(func)
    def inner(*args, **kwargs):
        warn_once(
            message="You're using a legacy version of Neptune client."
            " It will be moved to `neptune.legacy` as of `neptune-client==1.0.0`."
        )
        return func(*args, **kwargs)

    return inner



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "warn_once",
    "warn_about_unsupported_type",
    "NeptuneDeprecationWarning",
    "NeptuneWarning",
    "NeptuneUnsupportedType",
]

import os
import traceback
import warnings

import neptune


class NeptuneDeprecationWarning(DeprecationWarning):
    pass


class NeptuneWarning(Warning):
    pass


class NeptuneUnsupportedType(Warning):
    pass


warnings.simplefilter("always", category=NeptuneDeprecationWarning)

warned_once = set()
path_to_root_module = os.path.dirname(os.path.realpath(neptune.__file__))


def get_user_code_stack_level():
    call_stack = traceback.extract_stack()
    for level, stack_frame in enumerate(reversed(call_stack)):
        if path_to_root_module not in stack_frame.filename:
            return level
    return 2


def warn_once(message: str, *, exception: type(Exception) = None):
    if exception is None:
        exception = NeptuneDeprecationWarning

    if message not in warned_once:
        warnings.warn(
            message=message,
            category=exception,
            stacklevel=get_user_code_stack_level(),
        )
        warned_once.add(message)


def warn_about_unsupported_type(type_str: str):
    warn_once(
        message=f"""You're attempting to log a type that is not directly supported by Neptune ({type_str}).
        Convert the value to a supported type, such as a string or float, or use stringify_unsupported(obj)
        for dictionaries or collections that contain unsupported values.
        For more, see https://docs.neptune.ai/help/value_of_unsupported_type""",
        exception=NeptuneUnsupportedType,
    )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Table"]

import logging
from datetime import datetime
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)

from neptune.exceptions import MetadataInconsistency
from neptune.internal.backends.api_model import (
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.utils.paths import (
    join_paths,
    parse_path,
)
from neptune.internal.utils.run_state import RunState

logger = logging.getLogger(__name__)


class TableEntry:
    def __init__(
        self,
        backend: NeptuneBackend,
        container_type: ContainerType,
        _id: str,
        attributes: List[AttributeWithProperties],
    ):
        self._backend = backend
        self._container_type = container_type
        self._id = _id
        self._attributes = attributes

    def __getitem__(self, path: str) -> "LeaderboardHandler":
        return LeaderboardHandler(table_entry=self, path=path)

    def get_attribute_type(self, path: str) -> AttributeType:
        for attr in self._attributes:
            if attr.path == path:
                return attr.type
        raise ValueError("Could not find {} attribute".format(path))

    def get_attribute_value(self, path: str) -> Any:
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.RUN_STATE:
                    return RunState.from_api(attr.properties.value).value
                if _type in (
                    AttributeType.FLOAT,
                    AttributeType.INT,
                    AttributeType.BOOL,
                    AttributeType.STRING,
                    AttributeType.DATETIME,
                ):
                    return attr.properties.value
                if _type == AttributeType.FLOAT_SERIES or _type == AttributeType.STRING_SERIES:
                    return attr.properties.last
                if _type == AttributeType.IMAGE_SERIES:
                    raise MetadataInconsistency("Cannot get value for image series.")
                if _type == AttributeType.FILE:
                    raise MetadataInconsistency("Cannot get value for file attribute. Use download() instead.")
                if _type == AttributeType.FILE_SET:
                    raise MetadataInconsistency("Cannot get value for file set attribute. Use download() instead.")
                if _type == AttributeType.STRING_SET:
                    return set(attr.properties.values)
                if _type == AttributeType.GIT_REF:
                    return attr.properties.commit.commitId
                if _type == AttributeType.NOTEBOOK_REF:
                    return attr.properties.notebookName
                if _type == AttributeType.ARTIFACT:
                    return attr.properties.hash
                logger.error(
                    "Attribute type %s not supported in this version, yielding None. Recommended client upgrade.",
                    _type,
                )
                return None
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_attribute(self, path: str, destination: Optional[str]):
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.FILE:
                    self._backend.download_file(
                        container_id=self._id,
                        container_type=self._container_type,
                        path=parse_path(path),
                        destination=destination,
                    )
                    return
                raise MetadataInconsistency("Cannot download file from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_set_attribute(self, path: str, destination: Optional[str]):
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.FILE_SET:
                    self._backend.download_file_set(
                        container_id=self._id,
                        container_type=self._container_type,
                        path=parse_path(path),
                        destination=destination,
                    )
                    return
                raise MetadataInconsistency("Cannot download ZIP archive from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))


class LeaderboardHandler:
    def __init__(self, table_entry: TableEntry, path: str):
        self._table_entry = table_entry
        self._path = path

    def __getitem__(self, path: str) -> "LeaderboardHandler":
        return LeaderboardHandler(table_entry=self._table_entry, path=join_paths(self._path, path))

    def get(self):
        return self._table_entry.get_attribute_value(path=self._path)

    def download(self, destination: Optional[str]):
        attr_type = self._table_entry.get_attribute_type(self._path)
        if attr_type == AttributeType.FILE:
            return self._table_entry.download_file_attribute(self._path, destination)
        elif attr_type == AttributeType.FILE_SET:
            return self._table_entry.download_file_set_attribute(path=self._path, destination=destination)
        raise MetadataInconsistency("Cannot download file from attribute of type {}".format(attr_type))


class Table:
    def __init__(
        self,
        backend: NeptuneBackend,
        container_type: ContainerType,
        entries: List[LeaderboardEntry],
    ):
        self._backend = backend
        self._entries = entries
        self._container_type = container_type

    def to_rows(self) -> List[TableEntry]:
        return [
            TableEntry(
                backend=self._backend,
                container_type=self._container_type,
                _id=e.id,
                attributes=e.attributes,
            )
            for e in self._entries
        ]

    def to_pandas(self):
        import pandas as pd

        def make_attribute_value(
            attribute: AttributeWithProperties,
        ) -> Optional[Union[str, float, datetime]]:
            _type = attribute.type
            _properties = attribute.properties
            if _type == AttributeType.RUN_STATE:
                return RunState.from_api(_properties.value).value
            if _type in (
                AttributeType.FLOAT,
                AttributeType.INT,
                AttributeType.BOOL,
                AttributeType.STRING,
                AttributeType.DATETIME,
            ):
                return _properties.value
            if _type == AttributeType.FLOAT_SERIES or _type == AttributeType.STRING_SERIES:
                return _properties.last
            if _type == AttributeType.IMAGE_SERIES:
                return None
            if _type == AttributeType.FILE or _type == AttributeType.FILE_SET:
                return None
            if _type == AttributeType.STRING_SET:
                return ",".join(_properties.values)
            if _type == AttributeType.GIT_REF:
                return _properties.commit.commitId
            if _type == AttributeType.NOTEBOOK_REF:
                return _properties.notebookName
            if _type == AttributeType.ARTIFACT:
                return _properties.hash
            logger.error(
                "Attribute type %s not supported in this version, yielding None. Recommended client upgrade.",
                _type,
            )
            return None

        def make_row(
            entry: LeaderboardEntry,
        ) -> Dict[str, Optional[Union[str, float, datetime]]]:
            row: Dict[str, Union[str, float, datetime]] = dict()
            for attr in entry.attributes:
                value = make_attribute_value(attr)
                if value is not None:
                    row[attr.path] = value
            return row

        def sort_key(attr):
            domain = attr.split("/")[0]
            if domain == "sys":
                return 0, attr
            if domain == "monitoring":
                return 2, attr
            return 1, attr

        rows = dict((n, make_row(entry)) for (n, entry) in enumerate(self._entries))

        df = pd.DataFrame.from_dict(data=rows, orient="index")
        df = df.reindex(sorted(df.columns, key=sort_key), axis="columns")
        return df



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
import math
import os
import re
import sys
import time
from collections import namedtuple
from http.client import NOT_FOUND
from io import StringIO
from itertools import groupby
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
)

import requests
import six
from bravado.exception import HTTPNotFound

import neptune.exceptions as alpha_exceptions
from neptune.attributes import constants as alpha_consts
from neptune.attributes.constants import (
    MONITORING_TRACEBACK_ATTRIBUTE_PATH,
    SYSTEM_FAILED_ATTRIBUTE_PATH,
)
from neptune.common import exceptions as common_exceptions
from neptune.common.exceptions import ClientHttpError
from neptune.common.storage.storage_utils import normalize_file_name
from neptune.common.utils import (
    NoopObject,
    assure_directory_exists,
)
from neptune.internal import operation as alpha_operation
from neptune.internal.backends import hosted_file_operations as alpha_hosted_file_operations
from neptune.internal.backends.api_model import AttributeType
from neptune.internal.backends.operation_api_name_visitor import OperationApiNameVisitor as AlphaOperationApiNameVisitor
from neptune.internal.backends.operation_api_object_converter import (
    OperationApiObjectConverter as AlphaOperationApiObjectConverter,
)
from neptune.internal.backends.utils import handle_server_raw_response_messages
from neptune.internal.operation import (
    AssignBool,
    AssignString,
    ConfigFloatSeries,
    LogFloats,
    LogStrings,
)
from neptune.internal.utils import (
    base64_decode,
    base64_encode,
)
from neptune.internal.utils import paths as alpha_path_utils
from neptune.internal.utils.paths import parse_path
from neptune.legacy.api_exceptions import (
    ExperimentNotFound,
    ExperimentOperationErrors,
    NotebookNotFound,
    PathInExperimentNotFound,
    ProjectNotFound,
)
from neptune.legacy.backend import LeaderboardApiClient
from neptune.legacy.checkpoint import Checkpoint
from neptune.legacy.exceptions import (
    DeleteArtifactUnsupportedInAlphaException,
    DownloadArtifactsUnsupportedException,
    DownloadArtifactUnsupportedException,
    DownloadSourcesException,
    FileNotFound,
    NeptuneException,
)
from neptune.legacy.experiments import Experiment
from neptune.legacy.internal.api_clients.hosted_api_clients.mixins import HostedNeptuneMixin
from neptune.legacy.internal.api_clients.hosted_api_clients.utils import legacy_with_api_exceptions_handler
from neptune.legacy.internal.channels.channels import (
    ChannelNamespace,
    ChannelType,
    ChannelValueType,
)
from neptune.legacy.internal.utils.alpha_integration import (
    AlphaChannelDTO,
    AlphaChannelWithValueDTO,
    AlphaParameterDTO,
    AlphaPropertyDTO,
    channel_type_to_clear_operation,
    channel_type_to_operation,
    channel_value_type_to_operation,
    deprecated_img_to_alpha_image,
)
from neptune.legacy.internal.websockets.reconnecting_websocket_factory import ReconnectingWebsocketFactory
from neptune.legacy.model import (
    ChannelWithLastValue,
    LeaderboardEntry,
)
from neptune.legacy.notebook import Notebook

_logger = logging.getLogger(__name__)

LegacyExperiment = namedtuple(
    "LegacyExperiment",
    "shortId "
    "name "
    "timeOfCreation "
    "timeOfCompletion "
    "runningTime "
    "owner "
    "storageSize "
    "channelsSize "
    "tags "
    "description "
    "hostname "
    "state "
    "properties "
    "parameters",
)

LegacyLeaderboardEntry = namedtuple(
    "LegacyExperiment",
    "id "
    "organizationName "
    "projectName "
    "shortId "
    "name "
    "state "
    "timeOfCreation "
    "timeOfCompletion "
    "runningTime "
    "owner "
    "size "
    "tags "
    "description "
    "channelsLastValues "
    "parameters "
    "properties",
)

if TYPE_CHECKING:
    from neptune.legacy.internal.api_clients import HostedNeptuneBackendApiClient


class HostedAlphaLeaderboardApiClient(HostedNeptuneMixin, LeaderboardApiClient):
    @legacy_with_api_exceptions_handler
    def __init__(self, backend_api_client: "HostedNeptuneBackendApiClient"):
        self._backend_api_client = backend_api_client

        self._client_config = self._create_client_config(
            api_token=self.credentials.api_token, backend_client=self.backend_client
        )

        self.leaderboard_swagger_client = self._get_swagger_client(
            "{}/api/leaderboard/swagger.json".format(self._client_config.api_url),
            self._backend_api_client.http_client,
        )

        if sys.version_info >= (3, 7):
            try:
                os.register_at_fork(after_in_child=self._handle_fork_in_child)
            except AttributeError:
                pass

    def _handle_fork_in_child(self):
        self.leaderboard_swagger_client = NoopObject()

    @property
    def http_client(self):
        return self._backend_api_client.http_client

    @property
    def backend_client(self):
        return self._backend_api_client.backend_client

    @property
    def authenticator(self):
        return self._backend_api_client.authenticator

    @property
    def credentials(self):
        return self._backend_api_client.credentials

    @property
    def backend_swagger_client(self):
        return self._backend_api_client.backend_swagger_client

    @property
    def client_lib_version(self):
        return self._backend_api_client.client_lib_version

    @property
    def api_address(self):
        return self._client_config.api_url

    @property
    def display_address(self):
        return self._backend_api_client.display_address

    @property
    def proxies(self):
        return self._backend_api_client.proxies

    @legacy_with_api_exceptions_handler
    def get_project_members(self, project_identifier):
        try:
            r = self.backend_swagger_client.api.listProjectMembers(projectIdentifier=project_identifier).response()
            return r.result
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier)

    @legacy_with_api_exceptions_handler
    def create_experiment(
        self,
        project,
        name,
        description,
        params,
        properties,
        tags,
        abortable,  # deprecated in alpha
        monitored,  # deprecated in alpha
        git_info,
        hostname,
        entrypoint,
        notebook_id,
        checkpoint_id,
    ):
        if not isinstance(name, six.string_types):
            raise ValueError("Invalid name {}, should be a string.".format(name))
        if not isinstance(description, six.string_types):
            raise ValueError("Invalid description {}, should be a string.".format(description))
        if not isinstance(params, dict):
            raise ValueError("Invalid params {}, should be a dict.".format(params))
        if not isinstance(properties, dict):
            raise ValueError("Invalid properties {}, should be a dict.".format(properties))
        if hostname is not None and not isinstance(hostname, six.string_types):
            raise ValueError("Invalid hostname {}, should be a string.".format(hostname))
        if entrypoint is not None and not isinstance(entrypoint, six.string_types):
            raise ValueError("Invalid entrypoint {}, should be a string.".format(entrypoint))

        git_info = (
            {
                "commit": {
                    "commitId": git_info.commit_id,
                    "message": git_info.message,
                    "authorName": git_info.author_name,
                    "authorEmail": git_info.author_email,
                    "commitDate": git_info.commit_date,
                },
                "repositoryDirty": git_info.repository_dirty,
                "currentBranch": git_info.active_branch,
                "remotes": git_info.remote_urls,
            }
            if git_info
            else None
        )

        api_params = {
            "notebookId": notebook_id,
            "checkpointId": checkpoint_id,
            "projectIdentifier": str(project.internal_id),
            "cliVersion": self.client_lib_version,
            "gitInfo": git_info,
            "customId": None,
        }

        kwargs = {
            "experimentCreationParams": api_params,
            "X-Neptune-CliVersion": self.client_lib_version,
            "_request_options": {"headers": {"X-Neptune-LegacyClient": "true"}},
        }

        try:
            api_experiment = self.leaderboard_swagger_client.api.createExperiment(**kwargs).response().result
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

        experiment = self._convert_to_experiment(api_experiment, project)
        # Initialize new experiment
        init_experiment_operations = self._get_init_experiment_operations(
            name, description, params, properties, tags, hostname, entrypoint
        )
        self._execute_operations(
            experiment=experiment,
            operations=init_experiment_operations,
        )
        return experiment

    def upload_source_code(self, experiment, source_target_pairs):
        dest_path = alpha_path_utils.parse_path(alpha_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH)
        file_globs = [source_path for source_path, target_path in source_target_pairs]
        upload_files_operation = alpha_operation.UploadFileSet(
            path=dest_path,
            file_globs=file_globs,
            reset=True,
        )
        self._execute_upload_operations_with_400_retry(experiment, upload_files_operation)

    @legacy_with_api_exceptions_handler
    def get_notebook(self, project, notebook_id):
        try:
            api_notebook_list = (
                self.leaderboard_swagger_client.api.listNotebooks(
                    projectIdentifier=project.internal_id, id=[notebook_id]
                )
                .response()
                .result
            )

            if not api_notebook_list.entries:
                raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

            api_notebook = api_notebook_list.entries[0]

            return Notebook(
                backend=self,
                project=project,
                _id=api_notebook.id,
                owner=api_notebook.owner,
            )
        except HTTPNotFound:
            raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

    @legacy_with_api_exceptions_handler
    def get_last_checkpoint(self, project, notebook_id):
        try:
            api_checkpoint_list = (
                self.leaderboard_swagger_client.api.listCheckpoints(notebookId=notebook_id, offset=0, limit=1)
                .response()
                .result
            )

            if not api_checkpoint_list.entries:
                raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

            checkpoint = api_checkpoint_list.entries[0]
            return Checkpoint(checkpoint.id, checkpoint.name, checkpoint.path)
        except HTTPNotFound:
            raise NotebookNotFound(notebook_id=notebook_id, project=project.full_id)

    @legacy_with_api_exceptions_handler
    def create_notebook(self, project):
        try:
            api_notebook = (
                self.leaderboard_swagger_client.api.createNotebook(projectIdentifier=project.internal_id)
                .response()
                .result
            )

            return Notebook(
                backend=self,
                project=project,
                _id=api_notebook.id,
                owner=api_notebook.owner,
            )
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

    @legacy_with_api_exceptions_handler
    def create_checkpoint(self, notebook_id, jupyter_path, _file=None):
        if _file is not None:
            with self._upload_raw_data(
                api_method=self.leaderboard_swagger_client.api.createCheckpoint,
                data=_file,
                headers={"Content-Type": "application/octet-stream"},
                path_params={"notebookId": notebook_id},
                query_params={"jupyterPath": jupyter_path},
            ) as response:
                if response.status_code == NOT_FOUND:
                    raise NotebookNotFound(notebook_id=notebook_id)
                else:
                    response.raise_for_status()
                    CheckpointDTO = self.leaderboard_swagger_client.get_model("CheckpointDTO")
                    return CheckpointDTO.unmarshal(response.json())
        else:
            try:
                NewCheckpointDTO = self.leaderboard_swagger_client.get_model("NewCheckpointDTO")
                return (
                    self.leaderboard_swagger_client.api.createEmptyCheckpoint(
                        notebookId=notebook_id,
                        checkpoint=NewCheckpointDTO(path=jupyter_path),
                    )
                    .response()
                    .result
                )
            except HTTPNotFound:
                return None

    @legacy_with_api_exceptions_handler
    def get_experiment(self, experiment_id):
        api_attributes = self._get_api_experiment_attributes(experiment_id)
        attributes = api_attributes.attributes
        system_attributes = api_attributes.systemAttributes

        return LegacyExperiment(
            shortId=system_attributes.shortId.value,
            name=system_attributes.name.value,
            timeOfCreation=system_attributes.creationTime.value,
            timeOfCompletion=None,
            runningTime=system_attributes.runningTime.value,
            owner=system_attributes.owner.value,
            storageSize=system_attributes.size.value,
            channelsSize=0,
            tags=system_attributes.tags.values,
            description=system_attributes.description.value,
            hostname=system_attributes.hostname.value if system_attributes.hostname else None,
            state="running" if system_attributes.state.value == "running" else "succeeded",
            properties=[AlphaPropertyDTO(attr) for attr in attributes if AlphaPropertyDTO.is_valid_attribute(attr)],
            parameters=[AlphaParameterDTO(attr) for attr in attributes if AlphaParameterDTO.is_valid_attribute(attr)],
        )

    @legacy_with_api_exceptions_handler
    def set_property(self, experiment, key, value):
        """Save attribute casted to string under `alpha_consts.PROPERTIES_ATTRIBUTE_SPACE` namespace"""
        self._execute_operations(
            experiment=experiment,
            operations=[
                alpha_operation.AssignString(
                    path=alpha_path_utils.parse_path(f"{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{key}"),
                    value=str(value),
                )
            ],
        )

    @legacy_with_api_exceptions_handler
    def remove_property(self, experiment, key):
        self._remove_attribute(experiment, str_path=f"{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{key}")

    @legacy_with_api_exceptions_handler
    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        operations = [
            alpha_operation.AddStrings(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                values=tags_to_add,
            ),
            alpha_operation.RemoveStrings(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                values=tags_to_delete,
            ),
        ]
        self._execute_operations(
            experiment=experiment,
            operations=operations,
        )

    @staticmethod
    def _get_channel_attribute_path(channel_name: str, channel_namespace: ChannelNamespace) -> str:
        if channel_namespace == ChannelNamespace.USER:
            prefix = alpha_consts.LOG_ATTRIBUTE_SPACE
        else:
            prefix = alpha_consts.MONITORING_ATTRIBUTE_SPACE
        return f"{prefix}{channel_name}"

    def _create_channel(
        self,
        experiment: Experiment,
        channel_id: str,
        channel_name: str,
        channel_type: ChannelType,
        channel_namespace: ChannelNamespace,
    ):
        """This function is responsible for creating 'fake' channels in alpha projects.

        Since channels are abandoned in alpha api, we're mocking them using empty logging operation."""

        operation = channel_type_to_operation(channel_type)

        log_empty_operation = operation(
            path=alpha_path_utils.parse_path(self._get_channel_attribute_path(channel_name, channel_namespace)),
            values=[],
        )  # this operation is used to create empty attribute
        self._execute_operations(
            experiment=experiment,
            operations=[log_empty_operation],
        )
        return ChannelWithLastValue(
            AlphaChannelWithValueDTO(
                channelId=channel_id,
                channelName=channel_name,
                channelType=channel_type.value,
                x=None,
                y=None,
            )
        )

    @legacy_with_api_exceptions_handler
    def create_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        channel_id = f"{alpha_consts.LOG_ATTRIBUTE_SPACE}{name}"
        return self._create_channel(
            experiment,
            channel_id,
            channel_name=name,
            channel_type=ChannelType(channel_type),
            channel_namespace=ChannelNamespace.USER,
        )

    def _get_channels(self, experiment) -> List[AlphaChannelDTO]:
        try:
            return [
                AlphaChannelDTO(attr)
                for attr in self._get_attributes(experiment.internal_id)
                if AlphaChannelDTO.is_valid_attribute(attr)
            ]
        except HTTPNotFound:
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id,
            )

    @legacy_with_api_exceptions_handler
    def get_channels(self, experiment) -> Dict[str, AlphaChannelDTO]:
        api_channels = [
            channel
            for channel in self._get_channels(experiment)
            # return channels from LOG_ATTRIBUTE_SPACE namespace only
            if channel.id.startswith(alpha_consts.LOG_ATTRIBUTE_SPACE)
        ]
        return {ch.name: ch for ch in api_channels}

    @legacy_with_api_exceptions_handler
    def create_system_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        channel_id = f"{alpha_consts.MONITORING_ATTRIBUTE_SPACE}{name}"
        return self._create_channel(
            experiment,
            channel_id,
            channel_name=name,
            channel_type=ChannelType(channel_type),
            channel_namespace=ChannelNamespace.SYSTEM,
        )

    @legacy_with_api_exceptions_handler
    def get_system_channels(self, experiment) -> Dict[str, AlphaChannelDTO]:
        return {
            channel.name: channel
            for channel in self._get_channels(experiment)
            if (
                channel.channelType == ChannelType.TEXT.value
                and channel.id.startswith(alpha_consts.MONITORING_ATTRIBUTE_SPACE)
            )
        }

    @legacy_with_api_exceptions_handler
    def send_channels_values(self, experiment, channels_with_values):
        send_operations = []
        for channel_with_values in channels_with_values:
            channel_value_type = channel_with_values.channel_type
            operation = channel_value_type_to_operation(channel_value_type)

            if channel_value_type == ChannelValueType.IMAGE_VALUE:
                # IMAGE_VALUE requires minor data modification before it's sent
                data_transformer = deprecated_img_to_alpha_image
            else:
                # otherwise use identity function as transformer
                data_transformer = lambda e: e  # noqa: E731

            ch_values = [
                alpha_operation.LogSeriesValue(
                    value=data_transformer(ch_value.value),
                    step=ch_value.x,
                    ts=ch_value.ts,
                )
                for ch_value in channel_with_values.channel_values
            ]
            send_operations.append(
                operation(
                    path=alpha_path_utils.parse_path(
                        self._get_channel_attribute_path(
                            channel_with_values.channel_name,
                            channel_with_values.channel_namespace,
                        )
                    ),
                    values=ch_values,
                )
            )

        self._execute_operations(experiment, send_operations)

    def mark_failed(self, experiment, traceback):
        operations = []
        path = parse_path(SYSTEM_FAILED_ATTRIBUTE_PATH)
        traceback_values = [LogStrings.ValueType(val, step=None, ts=time.time()) for val in traceback.split("\n")]
        operations.append(AssignBool(path=path, value=True))
        operations.append(
            LogStrings(
                values=traceback_values,
                path=parse_path(MONITORING_TRACEBACK_ATTRIBUTE_PATH),
            )
        )
        self._execute_operations(experiment, operations)

    def ping_experiment(self, experiment):
        try:
            self.leaderboard_swagger_client.api.ping(experimentId=str(experiment.internal_id)).response().result
        except HTTPNotFound:
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id,
            )

    @staticmethod
    def _get_attribute_name_for_metric(resource_type, gauge_name, gauges_count) -> str:
        if gauges_count > 1:
            return "monitoring/{}_{}".format(resource_type, gauge_name).lower()
        return "monitoring/{}".format(resource_type).lower()

    @legacy_with_api_exceptions_handler
    def create_hardware_metric(self, experiment, metric):
        operations = []
        gauges_count = len(metric.gauges)
        for gauge in metric.gauges:
            path = parse_path(self._get_attribute_name_for_metric(metric.resource_type, gauge.name(), gauges_count))
            operations.append(ConfigFloatSeries(path, min=metric.min_value, max=metric.max_value, unit=metric.unit))
        self._execute_operations(experiment, operations)

    @legacy_with_api_exceptions_handler
    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        operations = []
        metrics_by_name = {metric.name: metric for metric in metrics}
        for report in metric_reports:
            metric = metrics_by_name.get(report.metric.name)
            gauges_count = len(metric.gauges)
            for gauge_name, metric_values in groupby(report.values, lambda value: value.gauge_name):
                metric_values = list(metric_values)
                path = parse_path(self._get_attribute_name_for_metric(metric.resource_type, gauge_name, gauges_count))
                operations.append(
                    LogFloats(
                        path,
                        [LogFloats.ValueType(value.value, step=None, ts=value.timestamp) for value in metric_values],
                    )
                )
        self._execute_operations(experiment, operations)

    def log_artifact(self, experiment, artifact, destination=None):
        if isinstance(artifact, str):
            if os.path.isfile(artifact):
                target_name = os.path.basename(artifact) if destination is None else destination
                dest_path = self._get_dest_and_ext(target_name)
                operation = alpha_operation.UploadFile(
                    path=dest_path,
                    ext="",
                    file_path=os.path.abspath(artifact),
                )
            elif os.path.isdir(artifact):
                for path, file_destination in self._log_dir_artifacts(artifact, destination):
                    self.log_artifact(experiment, path, file_destination)
                return
            else:
                raise FileNotFound(artifact)
        elif hasattr(artifact, "read"):
            if not destination:
                raise ValueError("destination is required for IO streams")
            dest_path = self._get_dest_and_ext(destination)
            data = artifact.read()
            content = data.encode("utf-8") if isinstance(data, str) else data
            operation = alpha_operation.UploadFileContent(path=dest_path, ext="", file_content=base64_encode(content))
        else:
            raise ValueError("Artifact must be a local path or an IO object")

        self._execute_upload_operations_with_400_retry(experiment, operation)

    @staticmethod
    def _get_dest_and_ext(target_name):
        qualified_target_name = f"{alpha_consts.ARTIFACT_ATTRIBUTE_SPACE}{target_name}"
        return alpha_path_utils.parse_path(normalize_file_name(qualified_target_name))

    def _log_dir_artifacts(self, directory_path, destination):
        directory_path = Path(directory_path)
        prefix = directory_path.name if destination is None else destination
        for path in directory_path.glob("**/*"):
            if path.is_file():
                relative_path = path.relative_to(directory_path)
                file_destination = prefix + "/" + str(relative_path)
                yield str(path), file_destination

    def delete_artifacts(self, experiment, path):
        try:
            self._remove_attribute(experiment, str_path=f"{alpha_consts.ARTIFACT_ATTRIBUTE_SPACE}{path}")
        except ExperimentOperationErrors as e:
            if all(isinstance(err, alpha_exceptions.MetadataInconsistency) for err in e.errors):
                raise DeleteArtifactUnsupportedInAlphaException(path, experiment) from None
            raise

    @legacy_with_api_exceptions_handler
    def download_data(self, experiment: Experiment, path: str, destination):
        project_storage_path = f"artifacts/{path}"
        with self._download_raw_data(
            api_method=self.leaderboard_swagger_client.api.downloadAttribute,
            headers={"Accept": "application/octet-stream"},
            path_params={},
            query_params={
                "experimentId": experiment.internal_id,
                "attribute": project_storage_path,
            },
        ) as response:
            if response.status_code == NOT_FOUND:
                raise PathInExperimentNotFound(path=path, exp_identifier=experiment.internal_id)
            else:
                response.raise_for_status()

            with open(destination, "wb") as f:
                for chunk in response.iter_content(chunk_size=10 * 1024 * 1024):
                    if chunk:
                        f.write(chunk)

    def download_sources(self, experiment: Experiment, path=None, destination_dir=None):
        if path is not None:
            # in alpha all source files stored as single FileSet must be downloaded at once
            raise DownloadSourcesException(experiment)
        path = alpha_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH

        destination_dir = assure_directory_exists(destination_dir)

        download_request = self._get_file_set_download_request(experiment.internal_id, path)
        alpha_hosted_file_operations.download_file_set_attribute(
            swagger_client=self.leaderboard_swagger_client,
            download_id=download_request.id,
            destination=destination_dir,
        )

    @legacy_with_api_exceptions_handler
    def _get_file_set_download_request(self, run_id: str, path: str):
        params = {
            "experimentId": run_id,
            "attribute": path,
        }
        return self.leaderboard_swagger_client.api.prepareForDownloadFileSetAttributeZip(**params).response().result

    def download_artifacts(self, experiment: Experiment, path=None, destination_dir=None):
        raise DownloadArtifactsUnsupportedException(experiment)

    def download_artifact(self, experiment: Experiment, path=None, destination_dir=None):
        destination_dir = assure_directory_exists(destination_dir)
        destination_path = os.path.join(destination_dir, os.path.basename(path))

        try:
            self.download_data(experiment, path, destination_path)
        except PathInExperimentNotFound:
            raise DownloadArtifactUnsupportedException(path, experiment) from None

    def _get_attributes(self, experiment_id) -> list:
        return self._get_api_experiment_attributes(experiment_id).attributes

    def _get_api_experiment_attributes(self, experiment_id):
        params = {
            "experimentId": experiment_id,
        }
        return self.leaderboard_swagger_client.api.getExperimentAttributes(**params).response().result

    def _remove_attribute(self, experiment, str_path: str):
        """Removes given attribute"""
        self._execute_operations(
            experiment=experiment,
            operations=[
                alpha_operation.DeleteAttribute(
                    path=alpha_path_utils.parse_path(str_path),
                )
            ],
        )

    @staticmethod
    def _get_client_config_args(api_token):
        return dict(
            X_Neptune_Api_Token=api_token,
            alpha="true",
        )

    def _execute_upload_operation(self, experiment: Experiment, upload_operation: alpha_operation.Operation):
        experiment_id = experiment.internal_id
        try:
            if isinstance(upload_operation, alpha_operation.UploadFile):
                alpha_hosted_file_operations.upload_file_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    container_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    source=upload_operation.file_path,
                    ext=upload_operation.ext,
                    multipart_config=self._client_config.multipart_config,
                )
            elif isinstance(upload_operation, alpha_operation.UploadFileContent):
                alpha_hosted_file_operations.upload_file_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    container_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    source=base64_decode(upload_operation.file_content),
                    ext=upload_operation.ext,
                    multipart_config=self._client_config.multipart_config,
                )
            elif isinstance(upload_operation, alpha_operation.UploadFileSet):
                alpha_hosted_file_operations.upload_file_set_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    container_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    file_globs=upload_operation.file_globs,
                    reset=upload_operation.reset,
                    multipart_config=self._client_config.multipart_config,
                )
            else:
                raise NeptuneException("Upload operation in neither File or FileSet")
        except common_exceptions.NeptuneException as e:
            raise NeptuneException(e) from e

        return None

    def _execute_upload_operations_with_400_retry(
        self, experiment: Experiment, upload_operation: alpha_operation.Operation
    ):
        while True:
            try:
                return self._execute_upload_operation(experiment, upload_operation)
            except ClientHttpError as ex:
                if "Length of stream does not match given range" not in ex.response:
                    raise ex

    @legacy_with_api_exceptions_handler
    def _execute_operations(self, experiment: Experiment, operations: List[alpha_operation.Operation]):
        experiment_id = experiment.internal_id
        file_operations = (
            alpha_operation.UploadFile,
            alpha_operation.UploadFileContent,
            alpha_operation.UploadFileSet,
        )
        if any(isinstance(op, file_operations) for op in operations):
            raise NeptuneException(
                "File operations must be handled directly by `_execute_upload_operation`,"
                " not by `_execute_operations` function call."
            )

        kwargs = {
            "experimentId": experiment_id,
            "operations": [
                {
                    "path": alpha_path_utils.path_to_str(op.path),
                    AlphaOperationApiNameVisitor().visit(op): AlphaOperationApiObjectConverter().convert(op),
                }
                for op in operations
            ],
        }
        try:
            result = self.leaderboard_swagger_client.api.executeOperations(**kwargs).response().result
            errors = [alpha_exceptions.MetadataInconsistency(err.errorDescription) for err in result]
            if errors:
                raise ExperimentOperationErrors(errors=errors)
            return None
        except HTTPNotFound as e:
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id,
            ) from e

    def _get_init_experiment_operations(
        self, name, description, params, properties, tags, hostname, entrypoint
    ) -> List[alpha_operation.Operation]:
        """Returns operations required to initialize newly created experiment"""
        init_operations = list()

        # Assign experiment name
        init_operations.append(
            alpha_operation.AssignString(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_NAME_ATTRIBUTE_PATH),
                value=name,
            )
        )
        # Assign experiment description
        init_operations.append(
            alpha_operation.AssignString(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_DESCRIPTION_ATTRIBUTE_PATH),
                value=description,
            )
        )
        # Assign experiment parameters
        for p_name, p_val in params.items():
            parameter_type, string_value = self._get_parameter_with_type(p_val)
            operation_cls = alpha_operation.AssignFloat if parameter_type == "double" else alpha_operation.AssignString
            init_operations.append(
                operation_cls(
                    path=alpha_path_utils.parse_path(f"{alpha_consts.PARAMETERS_ATTRIBUTE_SPACE}{p_name}"),
                    value=string_value,
                )
            )
        # Assign experiment properties
        for p_key, p_val in properties.items():
            init_operations.append(
                AssignString(
                    path=alpha_path_utils.parse_path(f"{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{p_key}"),
                    value=str(p_val),
                )
            )
        # Assign tags
        if tags:
            init_operations.append(
                alpha_operation.AddStrings(
                    path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                    values=set(tags),
                )
            )
        # Assign source hostname
        if hostname:
            init_operations.append(
                alpha_operation.AssignString(
                    path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_HOSTNAME_ATTRIBUTE_PATH),
                    value=hostname,
                )
            )
        # Assign source entrypoint
        if entrypoint:
            init_operations.append(
                alpha_operation.AssignString(
                    path=alpha_path_utils.parse_path(alpha_consts.SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH),
                    value=entrypoint,
                )
            )

        return init_operations

    @legacy_with_api_exceptions_handler
    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        op = channel_type_to_clear_operation(ChannelType(channel_type))
        attr_path = self._get_channel_attribute_path(channel_name, ChannelNamespace.USER)
        self._execute_operations(
            experiment=experiment,
            operations=[op(path=alpha_path_utils.parse_path(attr_path))],
        )

    @legacy_with_api_exceptions_handler
    def _get_channel_tuples_from_csv(self, experiment, channel_attribute_path):
        try:
            csv = (
                self.leaderboard_swagger_client.api.getFloatSeriesValuesCSV(
                    experimentId=experiment.internal_id,
                    attribute=channel_attribute_path,
                )
                .response()
                .incoming_response.text
            )
            lines = csv.split("\n")[:-1]
            return [line.split(",") for line in lines]
        except HTTPNotFound:
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id,
            )

    @legacy_with_api_exceptions_handler
    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        try:
            channel_attr_path = self._get_channel_attribute_path(channel_name, ChannelNamespace.USER)
            values = self._get_channel_tuples_from_csv(experiment, channel_attr_path)
            step_and_value = [val[0] + "," + val[2] for val in values]
            csv = StringIO()
            for line in step_and_value:
                csv.write(line + "\n")
            csv.seek(0)
            return csv
        except HTTPNotFound:
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id,
            )

    @legacy_with_api_exceptions_handler
    def get_metrics_csv(self, experiment):
        metric_channels = [
            channel
            for channel in self._get_channels(experiment)
            if (
                channel.channelType == ChannelType.NUMERIC.value
                and channel.id.startswith(alpha_consts.MONITORING_ATTRIBUTE_SPACE)
            )
        ]
        data = {
            # val[1] + ',' + val[2] is timestamp,value
            ch.name: [val[1] + "," + val[2] for val in self._get_channel_tuples_from_csv(experiment, ch.id)]
            for ch in metric_channels
        }
        values_count = max(len(values) for values in data.values())
        csv = StringIO()
        csv.write(",".join(["x_{name},y_{name}".format(name=ch.name) for ch in metric_channels]))
        csv.write("\n")
        for i in range(0, values_count):
            csv.write(",".join([data[ch.name][i] if i < len(data[ch.name]) else "," for ch in metric_channels]))
            csv.write("\n")
        csv.seek(0)
        return csv

    @legacy_with_api_exceptions_handler
    def get_leaderboard_entries(
        self,
        project,
        entry_types=None,  # deprecated
        ids=None,
        states=None,
        owners=None,
        tags=None,
        min_running_time=None,
    ):
        if states is not None:
            states = [state if state == "running" else "idle" for state in states]
        try:

            def get_portion(limit, offset):
                return (
                    self.leaderboard_swagger_client.api.getLeaderboard(
                        projectIdentifier=project.full_id,
                        shortId=ids,
                        state=states,
                        owner=owners,
                        tags=tags,
                        tagsMode="and",
                        minRunningTimeSeconds=min_running_time,
                        sortBy=["sys/id"],
                        sortFieldType=["string"],
                        sortDirection=["ascending"],
                        limit=limit,
                        offset=offset,
                    )
                    .response()
                    .result.entries
                )

            return [
                LeaderboardEntry(self._to_leaderboard_entry_dto(e)) for e in self._get_all_items(get_portion, step=100)
            ]
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

    def websockets_factory(self, project_id, experiment_id):
        base_url = re.sub(r"^http", "ws", self.api_address) + "/api/notifications/v1"
        return ReconnectingWebsocketFactory(backend=self, url=base_url + f"/runs/{project_id}/{experiment_id}/signal")

    @staticmethod
    def _to_leaderboard_entry_dto(experiment_attributes):
        attributes = experiment_attributes.attributes
        system_attributes = experiment_attributes.systemAttributes

        def is_channel_namespace(name):
            return name.startswith(alpha_consts.LOG_ATTRIBUTE_SPACE) or name.startswith(
                alpha_consts.MONITORING_ATTRIBUTE_SPACE
            )

        numeric_channels = [
            HostedAlphaLeaderboardApiClient._float_series_to_channel_last_value_dto(attr)
            for attr in attributes
            if (
                attr.type == AttributeType.FLOAT_SERIES.value
                and is_channel_namespace(attr.name)
                and attr.floatSeriesProperties.last is not None
            )
        ]
        text_channels = [
            HostedAlphaLeaderboardApiClient._string_series_to_channel_last_value_dto(attr)
            for attr in attributes
            if (
                attr.type == AttributeType.STRING_SERIES.value
                and is_channel_namespace(attr.name)
                and attr.stringSeriesProperties.last is not None
            )
        ]

        return LegacyLeaderboardEntry(
            id=experiment_attributes.id,
            organizationName=experiment_attributes.organizationName,
            projectName=experiment_attributes.projectName,
            shortId=system_attributes.shortId.value,
            name=system_attributes.name.value,
            state="running" if system_attributes.state.value == "running" else "succeeded",
            timeOfCreation=system_attributes.creationTime.value,
            timeOfCompletion=None,
            runningTime=system_attributes.runningTime.value,
            owner=system_attributes.owner.value,
            size=system_attributes.size.value,
            tags=system_attributes.tags.values,
            description=system_attributes.description.value,
            channelsLastValues=numeric_channels + text_channels,
            parameters=[AlphaParameterDTO(attr) for attr in attributes if AlphaParameterDTO.is_valid_attribute(attr)],
            properties=[AlphaPropertyDTO(attr) for attr in attributes if AlphaPropertyDTO.is_valid_attribute(attr)],
        )

    @staticmethod
    def _float_series_to_channel_last_value_dto(attribute):
        return AlphaChannelWithValueDTO(
            channelId=attribute.name,
            channelName=attribute.name.split("/", 1)[-1],
            channelType="numeric",
            x=attribute.floatSeriesProperties.lastStep,
            y=attribute.floatSeriesProperties.last,
        )

    @staticmethod
    def _string_series_to_channel_last_value_dto(attribute):
        return AlphaChannelWithValueDTO(
            channelId=attribute.name,
            channelName=attribute.name.split("/", 1)[-1],
            channelType="text",
            x=attribute.stringSeriesProperties.lastStep,
            y=attribute.stringSeriesProperties.last,
        )

    @staticmethod
    def _get_all_items(get_portion, step):
        items = []

        previous_items = None
        while previous_items is None or len(previous_items) >= step:
            previous_items = get_portion(limit=step, offset=len(items))
            items += previous_items

        return items

    def _upload_raw_data(self, api_method, data, headers, path_params, query_params):
        url = self.api_address + api_method.operation.path_name + "?"

        for key, val in path_params.items():
            url = url.replace("{" + key + "}", val)

        for key, val in query_params.items():
            url = url + key + "=" + val + "&"

        session = self.http_client.session

        request = self.authenticator.apply(requests.Request(method="POST", url=url, data=data, headers=headers))

        return handle_server_raw_response_messages(session.send(session.prepare_request(request)))

    def _get_parameter_with_type(self, parameter):
        string_type = "string"
        double_type = "double"
        if isinstance(parameter, bool):
            return (string_type, str(parameter))
        elif isinstance(parameter, float) or isinstance(parameter, int):
            if math.isinf(parameter) or math.isnan(parameter):
                return (string_type, json.dumps(parameter))
            else:
                return (double_type, str(parameter))
        else:
            return (string_type, str(parameter))

    def _convert_to_experiment(self, api_experiment, project):
        return Experiment(
            backend=project._backend,
            project=project,
            _id=api_experiment.shortId,
            internal_id=api_experiment.id,
        )

    def _download_raw_data(self, api_method, headers, path_params, query_params):
        url = self.api_address + api_method.operation.path_name + "?"

        for key, val in path_params.items():
            url = url.replace("{" + key + "}", val)

        for key, val in query_params.items():
            url = url + key + "=" + val + "&"

        session = self.http_client.session

        request = self.authenticator.apply(requests.Request(method="GET", url=url, headers=headers))

        return handle_server_raw_response_messages(session.send(session.prepare_request(request), stream=True))
